﻿//******************************************************************************
//* Simple assembler for the MiniRISC CPU.                                     *
//*                                                                            *
//* Written by: Tamas Raikovich                                                *
//* Version   : 2.0                                                            *
//******************************************************************************
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace MiniRISC_as
{
    /// <summary>
    /// MiniRISC assembler.
    /// </summary>
    public class Assembler : IDisposable
    {
        #region Enumerations.

        /// <summary>
        /// Instruction opcodes.
        /// </summary>
        private enum Opcode
        {
            //r : register (R0-R15)
            //p : program memory address (0x00-0xff)
            //d : data memory address (0x00-0xff)
            //i : indirect address from register (R0-R15)
            //#c: constant (0x00-0xff)

            //MOV Rd, d instruction (load, absolute addressing).
            MOV_R_D = 0,
            //MOV Rd, d instruction (load, indirect addressing).
            MOV_R_I,
            //MOV d, Ra instruction (store, absolute addressing).
            MOV_D_R,
            //MOV d, Ra instruction (store, indirect addressing).
            MOV_I_R,
            //MOV Rd, #c instruction (load constant).
            MOV_R_C,
            //ADD Rd, Ra, Rb instruction (Rd = Ra + Rb).
            ADD_R_R_R,
            //ADD Rd, #c instruction (Rd = Rd + c).
            ADD_R_C,
            //SUB Rd, Ra, Rb instruction (Rd = Ra - Rb).
            SUB_R_R_R,
            //AND Rd, Ra, Rb instruction (Rd = Ra & Rb).
            AND_R_R_R,
            //AND Rd, #c instruction (Rd = Rd & c).
            AND_R_C,
            //OR Rd, Ra, Rb instruction (Rd = Ra | Rb).
            OR_R_R_R,
            //OR Rd, #c instruction (Rd = Rd | c).
            OR_R_C,
            //XOR Rd, Ra, Rb instruction (Rd = Ra ^ Rb).
            XOR_R_R_R,
            //XOR Rd, #c instruction (Rd = Rd ^ c).
            XOR_R_C,
            //Shift instructions.
            SHIFT,
            //Program control instructions.
            PRG_CTRL
        }

        /// <summary>
        /// Shift instructions.
        /// </summary>
        private enum ShiftType
        {
            SL0 = 0x00,
            SR0 = 0x01,
            SL1 = 0x02,
            SR1 = 0x03,
            ROL = 0x04,
            ROR = 0x05
        }

        /// <summary>
        /// Program control instructions.
        /// </summary>
        private enum PrgCtrlType
        {
            JMP = 0x00,
            JZ  = 0x01,
            JNZ = 0x02,
            JC  = 0x03,
            JNC = 0x04,
            JSR = 0x05,
            RTS = 0x06
        }

        #endregion

        //Name of the source file with and without the extension.
        private string sourceFileNameWithExt;
        private string sourceFileName;

        //Current line number.
        private int lineNumber;
        //Current source code line.
        private string currentLine;

        //Symbol table dictionary.
        private Dictionary<string, SymbolTableItem> symbolTable = new Dictionary<string, SymbolTableItem>();
        //List for storing the unresolved symbols.
        private List<UnresolvedSymbol> unresolvedSymbols = new List<UnresolvedSymbol>();

        //Array that stores the program memory data.
        private int[] programMemory = new int[256];
        //Current address int he program memory.
        private int programMemoryAddress;

        //LexicalAnalyzer object.
        private LexicalAnalyzer lexicalAnalyzer = new LexicalAnalyzer();

        //Maximum number of errors before stop.
        private int maxError = 10;
        //Number of errors.
        private int numOfErrors;


        #region Events.

        /// <summary>
        /// Message event handler.
        /// </summary>
        /// <param name="sender">Object that raised the event.</param>
        /// <param name="msg">String that contains the message.</param>
        public delegate void AssemblerMsgEventHandler(object sender, string msg);

        /// <summary>
        /// Message event.
        /// </summary>
        public event AssemblerMsgEventHandler Message;

        #endregion

        #region Properties.

        /// <summary>
        /// Gets or sets the number of errors before stop.
        /// </summary>
        public int MaxError
        {
            get
            {
                return maxError;
            }
            set
            {
                if (value <= 0)
                    throw new ArgumentException("Value must be positive.", "MaxError");
                maxError = value;
            }
        }

        /// <summary>
        /// Gets the number of errors.
        /// </summary>
        public int NumOfErrors { get { return numOfErrors; } }

        #endregion


        #region Methods for opening and closing the assembly source file.

        /// <summary>
        /// Opens the assembly source file.
        /// </summary>
        /// <param name="filename">Name of the file to open.</param>
        public void Open(string filename)
        {
            //Open the source file with the lexical analyzer.
            lexicalAnalyzer.Open(filename);

            //Set the number of errors to zero.
            numOfErrors = 0;

            //Clear the symbol table and the unresolved labels list.
            symbolTable.Clear();
            unresolvedSymbols.Clear();

            //Clear the program memory data array.
            for (int i = 0; i < programMemory.Length; i++)
                programMemory[i] = 0;
            programMemoryAddress = 0;

            //Remove the extension from the source file name.
            int extensionPos = filename.Length;

            for (int i = filename.Length - 1; i >= 0; i--)
            {
                if (filename[i] == '.')
                {
                    extensionPos = i;
                    break;
                }
                if ((filename[i] == '\\') || (filename[i] == '/'))
                {
                    break;
                }
            }

            sourceFileName = filename.Substring(0, extensionPos);
            sourceFileNameWithExt = filename;
        }

        /// <summary>
        /// Closes the assembly source file.
        /// </summary>
        public void Close()
        {
            lexicalAnalyzer.Close();
        }

        #endregion

        #region IDisposable members.

        /// <summary>
        /// Releases the system resources associated with the Assembler object.
        /// </summary>
        void IDisposable.Dispose()
        {
            Close();
        }

        #endregion

        #region Methods for message and error handling.

        /// <summary>
        /// Raises an AsemblerException exception.
        /// </summary>
        /// <param name="msg">Error message.</param>
        private void ErrorMsg(string msg)
        {
            throw new AssemblerException("Error at line " + lineNumber.ToString() + ": " + msg);
        }

        /// <summary>
        /// Raises an AsemblerException exception.
        /// </summary>
        /// <param name="lineNumber">Line number.</param>
        /// <param name="msg">Error message.</param>
        private void ErrorMsg(int lineNumber, string msg)
        {
            throw new AssemblerException("Error at line " + lineNumber.ToString() + ": " + msg);
        }

        /// <summary>
        /// Raises the Message event with the error message.
        /// </summary>
        /// <param name="msg">Error message.</param>
        /// <returns>True, if the number of errors exceeds the maximum allowed value. False otherwise.</returns>
        private bool OnError(string msg)
        {
            //Increment the number of errors.
            numOfErrors++;

            //Raise the message event.
            OnMessage(msg);

            //Check the number of errors.
            if (numOfErrors >= maxError)
                return true;

            return false;
        }

        /// <summary>
        /// Raises the Message event.
        /// </summary>
        /// <param name="msg">Message.</param>
        private void OnMessage(string msg)
        {
            AssemblerMsgEventHandler eventHandler = Message;
            if (eventHandler != null)
                eventHandler(this, msg);
        }

        #endregion

        #region Methods for generating the output files.

        /// <summary>
        /// Generates the assembly listing file.
        /// </summary>
        private void GenerateListFile()
        {
            OnMessage("Generating the LST file...");

            int maxSymbolLength = 0;
            int maxIdentifierLength = 0;
            int maxLabelLength = 0;

            //Get the length of the longest symbol, identifier and label.
            foreach (string symbol in symbolTable.Keys)
            {
                SymbolTableItem item = symbolTable[symbol];

                if (symbol.Length > maxSymbolLength)
                    maxSymbolLength = symbol.Length;
                if ((item.SymbolType == SymbolType.Identifier) && (symbol.Length > maxIdentifierLength))
                    maxIdentifierLength = symbol.Length;
                if ((item.SymbolType == SymbolType.Label) && (symbol.Length > maxLabelLength))
                    maxLabelLength = symbol.Length;
            }

            //Length of the indents in the list file.
            int sourceCodeIndent = 14;
            int instructionIndent = 4;
            int operandsIndent = 8;
            int defValueIndent = maxIdentifierLength + 1;

            //Start position of the comments.
            int commentStartPos1 = sourceCodeIndent + instructionIndent + operandsIndent + maxSymbolLength + 4;
            int commentStartpos2 = sourceCodeIndent + instructionIndent + operandsIndent + 6 + maxIdentifierLength + 4;
            int commentStartpos3 = sourceCodeIndent + operandsIndent + defValueIndent + 4;
            int commentStartPos = sourceCodeIndent + instructionIndent + operandsIndent + 14;

            if (commentStartPos < commentStartpos2)
                commentStartPos = commentStartpos2;
            if (commentStartPos < commentStartpos3)
                commentStartPos = commentStartpos3;
            commentStartPos++;
            commentStartPos = (commentStartPos + 3) / 4;
            commentStartPos = commentStartPos * 4;


            using (StreamWriter listFile = new StreamWriter(sourceFileName + ".lst", false, Encoding.ASCII))
            {
                //Write the header to the list file.
                listFile.WriteLine("MiniRISC assembler v2.1 list file");
                listFile.WriteLine("Copyright (C) 2011 LOGSYS, Tamas Raikovich");
                listFile.WriteLine("Source file: " + sourceFileNameWithExt);
                listFile.WriteLine("Created on : " + DateTime.Now.ToString());
                listFile.WriteLine();
                listFile.WriteLine("Addr  Instr   Source code");
                for (int i = 0; i < 94; i++)
                    listFile.Write("-");
                listFile.WriteLine();

                //Reset the file pointer in the source file.
                lexicalAnalyzer.ResetFilePointer();

                //Program memory address.
                int address = 0;

                //Process the source file.
                for (; ; )
                {
                    //Set the character position in the current line to 0.
                    int pos = 0;

                    //Get the next symbol from the source file.
                    lexicalAnalyzer.GetNextSymbol();

                    //Exit from the loop if the end of file has been reached.
                    if (lexicalAnalyzer.SymbolType == SymbolType.EndOfFile)
                        break;

                    //Add a new line to the list file and get the next symbol if end of the line has been reached.
                    if (lexicalAnalyzer.SymbolType == SymbolType.EndOfLine)
                    {
                        listFile.WriteLine();
                        continue;
                    }

                    //Write the comment to the list file and get the next symbol if the symbol is a ';'.
                    if (lexicalAnalyzer.SymbolType == SymbolType.SemiColon)
                    {
                        for (int i = 0; i < sourceCodeIndent; i++)
                            listFile.Write(" ");
                        listFile.Write(lexicalAnalyzer.SymbolString);
                        listFile.WriteLine();
                        continue;
                    }

                    //Check whether the symbol is a label.
                    if (lexicalAnalyzer.SymbolType == SymbolType.Label)
                    {
                        string label = lexicalAnalyzer.SymbolString + ":";

                        //Write the address and the label to the list file.
                        listFile.Write(address.ToString("X2"));
                        for (int i = 0; i < (sourceCodeIndent - 2); i++)
                            listFile.Write(" ");
                        listFile.Write(label);

                        //Get the next symbol from the source file.
                        lexicalAnalyzer.GetNextSymbol();

                        //Write the comment to the list file and get the next symbol if the symbol is a ';'.
                        if (lexicalAnalyzer.SymbolType == SymbolType.SemiColon)
                        {
                            int indent = commentStartPos - sourceCodeIndent - label.Length;

                            for (int i = 0; i < indent; i++)
                                listFile.Write(" ");
                            listFile.Write(lexicalAnalyzer.SymbolString);
                            listFile.WriteLine();
                            continue;
                        }

                        //Add a new line to the list file after the label.
                        listFile.WriteLine();

                        //Get the next symbol if end of the line has been reached.
                        if (lexicalAnalyzer.SymbolType == SymbolType.EndOfLine)
                            continue;
                    }

                    bool isDirective = false;

                    //Process the directives and the instructions.
                    for (; ; )
                    {
                        int indent = 0;

                        //Check the symbol type.
                        switch (lexicalAnalyzer.SymbolType)
                        {
                            case SymbolType.Keyword:
                                //Check whether the symbol is a directive.
                                if (lexicalAnalyzer.IsDirective)
                                {
                                    isDirective = true;
                                    for (int i = 0; i < sourceCodeIndent; i++)
                                        listFile.Write(" ");
                                    pos += sourceCodeIndent;
                                }
                                //Check whether the symbol is an instruction.
                                if (lexicalAnalyzer.IsInstruction)
                                {
                                    //Write the program memory address and the instruction to the list file.
                                    listFile.Write(address.ToString("X2"));
                                    for (int i = 0; i < 4; i++)
                                        listFile.Write(" ");
                                    listFile.Write(programMemory[address].ToString("X4"));
                                    for (int i = 0; i < (4 + instructionIndent); i++)
                                        listFile.Write(" ");
                                    pos += sourceCodeIndent + instructionIndent;
                                    
                                    //Increment the program memory address.
                                    address++;
                                }
                                //Write the keyword to the list file.
                                if (lexicalAnalyzer.IsDirective)
                                    listFile.Write(lexicalAnalyzer.SymbolString.ToUpper());
                                else
                                    listFile.Write(lexicalAnalyzer.SymbolString.ToLower());
                                pos += lexicalAnalyzer.SymbolString.Length;
                                //Check whether the symbol is a directive.
                                if (lexicalAnalyzer.IsDirective)
                                {
                                    listFile.Write(" ");
                                    pos++;
                                }
                                //Check whether the symbol is an instruction.
                                if (lexicalAnalyzer.IsInstruction)
                                {
                                    indent = operandsIndent - lexicalAnalyzer.SymbolString.Length;
                                    for (int i = 0; i < indent; i++)
                                        listFile.Write(" ");
                                    pos += indent;
                                }
                                break;
                            case SymbolType.Identifier:
                                listFile.Write(lexicalAnalyzer.SymbolString);
                                pos += lexicalAnalyzer.SymbolString.Length;
                                if (isDirective == false)
                                {
                                    listFile.Write("[");
                                    listFile.Write(symbolTable[lexicalAnalyzer.SymbolString].SymbolValue.ToString("X2"));
                                    listFile.Write("]");
                                    pos += 4;
                                }
                                else
                                {
                                    indent = defValueIndent - lexicalAnalyzer.SymbolString.Length;
                                    for (int i = 0; i < indent; i++)
                                        listFile.Write(" ");
                                    pos += indent;
                                }
                                break;
                            case SymbolType.Number:
                                listFile.Write("0x");
                                listFile.Write(lexicalAnalyzer.SymbolValue.ToString("X2"));
                                pos += 4;
                                break;
                            case SymbolType.Constant:
                                listFile.Write("#0x");
                                listFile.Write(lexicalAnalyzer.SymbolValue.ToString("X2"));
                                pos += 5;
                                break;
                            case SymbolType.EndOfLine:
                                break;
                            case SymbolType.EndOfFile:
                                break;
                            case SymbolType.SemiColon:
                                indent = commentStartPos - pos;
                                for (int i = 0; i < indent; i++)
                                    listFile.Write(" ");
                                listFile.Write(lexicalAnalyzer.SymbolString);
                                break;
                            case SymbolType.Comma:
                                listFile.Write(", ");
                                pos += 2;
                                break;
                            default:
                                listFile.Write(lexicalAnalyzer.SymbolString);
                                pos += lexicalAnalyzer.SymbolString.Length;
                                break;
                        }

                        //Exit from the loop if the end of line/file has been reached.
                        if ((lexicalAnalyzer.SymbolType == SymbolType.EndOfFile) ||
                            (lexicalAnalyzer.SymbolType == SymbolType.EndOfLine) ||
                            (lexicalAnalyzer.SymbolType == SymbolType.SemiColon))
                        {
                            listFile.WriteLine();
                            break;
                        }

                        //Get the next symbol from the source file.
                        lexicalAnalyzer.GetNextSymbol();
                    }
                }
            }
        }

        /// <summary>
        /// Generates a HEX file that contains the program memory data.
        /// </summary>
        private void GenerateHexFile()
        {
            OnMessage("Generating the HEX file...");

            using (StreamWriter hexFile = new StreamWriter(sourceFileName + ".hex", false, Encoding.ASCII))
            {
                for (int i = 0; i < programMemory.Length; i++)
                {
                    hexFile.WriteLine(programMemory[i].ToString("X4"));
                }
            }
        }

        /// <summary>
        /// Generates an SVF file for updating the program memory through the JTAG interface.
        /// </summary>
        private void GenerateSvfFile()
        {
            OnMessage("Generating the SVF file...");

            using (StreamWriter svfFile = new StreamWriter(sourceFileName + ".svf", false, Encoding.ASCII))
            {
                svfFile.WriteLine("TRST OFF;");
                svfFile.WriteLine("ENDIR IDLE;");
                svfFile.WriteLine("ENDDR IDLE;");
                svfFile.WriteLine("STATE RESET IDLE;");
                svfFile.WriteLine("TDR 0 TDI() SMASK ();");
                svfFile.WriteLine("TIR 0 TDI() SMASK ();");
                svfFile.WriteLine("HDR 0 TDI() SMASK ();");
                svfFile.WriteLine("HIR 0 TDI() SMASK ();");
                svfFile.WriteLine("SIR 6 TDI (02) SMASK (3F);");

                for (int i = 0; i < programMemory.Length; i++)
                {
                    if (i >= programMemoryAddress)
                        break;

                    svfFile.Write("SDR 24 TDI (");
                    svfFile.Write(i.ToString("X2"));
                    svfFile.Write(programMemory[i].ToString("X4"));
                    svfFile.Write(") SMASK (FFFFFF);");
                    svfFile.WriteLine();
                }

                svfFile.WriteLine("SIR 6 TDI (3F) SMASK (3F);");
            }
        }

        #endregion

        #region Methods for writing the program memory.

        /// <summary>
        /// Writes an instruction with three operands to the program memory.
        /// </summary>
        /// <param name="opcode">Opcode.</param>
        /// <param name="rd">Destination register.</param>
        /// <param name="ra">Source register A.</param>
        /// <param name="rb">Source register B.</param>
        private void WriteProgramMemory(Opcode opcode, int rd, int ra, int rb)
        {
            //Check the program memory address.

            //Write the instruction to the prograsm memory.
            int instruction = (((int)opcode << 12) & 0xf000) | ((rd << 8) & 0x0f00) | ((ra << 4) & 0x00f0) | (rb & 0x000f);
            programMemory[programMemoryAddress] = instruction;

            //Increment the program memory address.
            programMemoryAddress++;
        }

        /// <summary>
        /// Writes an instruction with two operands to the program memory.
        /// </summary>
        /// <param name="opcode">Opcode.</param>
        /// <param name="reg">Register.</param>
        /// <param name="constant">Constant.</param>
        private void WriteProgramMemory(Opcode opcode, int reg, int constant)
        {
            //Check the program memory address.

            //Write the instruction to the prograsm memory.
            int instruction = (((int)opcode << 12) & 0xf000) | ((reg << 8) & 0x0f00) | (constant & 0x00ff);
            programMemory[programMemoryAddress] = instruction;

            //Increment the program memory address.
            programMemoryAddress++;
        }

        /// <summary>
        /// Updates the constant value in an instruction.
        /// </summary>
        /// <param name="unresolvedSymbol">Data associated with the unresolved label.</param>
        /// <param name="value">New jump address.</param>
        private void UpdateSymbolValue(UnresolvedSymbol unresolvedSymbol, int value)
        {
            int instruction = programMemory[unresolvedSymbol.ProgramMemoryAddress];

            //Update the constant value.
            instruction = (instruction & 0xff00) | (value & 0x00ff);
            programMemory[unresolvedSymbol.ProgramMemoryAddress] = instruction;
        }

        #endregion

        #region Methods for reading from the source file.

        /// <summary>
        /// Returns the register index if the next symbol is a register.
        /// </summary>
        /// <returns>Register index.</returns>
        private int GetRegister()
        {
            //The next symbol must be a keyword.
            if ((lexicalAnalyzer.SymbolType != SymbolType.Keyword) ||
                (lexicalAnalyzer.Keyword < Keyword.R0) ||
                (lexicalAnalyzer.Keyword > Keyword.R15))
            {
                if (lexicalAnalyzer.SymbolString.Length == 0)
                    ErrorMsg("register is expected.");
                else
                    ErrorMsg("register is expected but found '" + lexicalAnalyzer.SymbolString + "'.");
            }

            //Calculate the register index.
            int registerIndex = (int)lexicalAnalyzer.Keyword - (int)Keyword.R0;

            //Get the next symbol from the source file.
            lexicalAnalyzer.GetNextSymbol();

            return registerIndex;
        }

        /// <summary>
        /// Returns the value associated to an identifier.
        /// </summary>
        /// <param name="identifier">Name of the identifier.</param>
        /// <returns>Value associated to the identifier.</returns>
        private long GetIdentifierValue(string identifier)
        {
            SymbolTableItem item;
            long value = 0;

            //Check the symbol table.
            if (symbolTable.TryGetValue(identifier, out item))
            {
                //The identifier has been found in the symbol table.
                value = item.SymbolValue;
            }
            else
            {
                //Put the identifier name into the unresolved symbols list.
                unresolvedSymbols.Add(new UnresolvedSymbol(lineNumber, programMemoryAddress, identifier));
            }

            return value;
        }

        /// <summary>
        /// Tries to get a constant value from the source file.
        /// </summary>
        /// <param name="constant">Variable to store the constant.</param>
        /// <returns>True if a constant value has been successfully read.</returns>
        private bool TryGetConstant(out long constant)
        {
            //Check whether the current symbol is a constant or a '#'.
            if ((lexicalAnalyzer.SymbolType != SymbolType.Sharp) &&
                (lexicalAnalyzer.SymbolType != SymbolType.Constant))
            {
                constant = 0;
                return false;
            }

            //Get the constant.
            if (lexicalAnalyzer.SymbolType == SymbolType.Constant)
            {
                constant = lexicalAnalyzer.SymbolValue;
            }
            else
            {
                lexicalAnalyzer.GetNextSymbol(false);
                if (lexicalAnalyzer.SymbolType != SymbolType.Identifier)
                {
                    if (lexicalAnalyzer.SymbolString.Length == 0)
                        ErrorMsg("invalid identifier.");
                    else
                        ErrorMsg("invalid identifier '" + lexicalAnalyzer.SymbolString + "'.");
                }
                constant = GetIdentifierValue(lexicalAnalyzer.SymbolString);
            }

            //Get the next symbol from the source file.
            lexicalAnalyzer.GetNextSymbol();

            return true;
        }

        /// <summary>
        /// Gets a constant value from the source file.
        /// </summary>
        /// <returns></returns>
        private long GetConstant()
        {
            long constant = 0;

            if (TryGetConstant(out constant) == false)
            {
                if (lexicalAnalyzer.SymbolString.Length == 0)
                    ErrorMsg("numeric constant is expected.");
                else
                    ErrorMsg("numeric constant is expected but found '" + lexicalAnalyzer.SymbolString + "'.");
            }

            return constant;
        }

        /// <summary>
        /// Checks whether the next symbol is a ',' character.
        /// </summary>
        private void CheckComma()
        {
            //Check the current symbol.
            if (lexicalAnalyzer.SymbolType != SymbolType.Comma)
                ErrorMsg("',' is expected.");

            //Get the next symbol from the source file.
            lexicalAnalyzer.GetNextSymbol();
        }

        /// <summary>
        /// Checks whether the next symbol is a ')' character.
        /// </summary>
        private void CheckRParenthesis()
        {
            //Check the current symbol.
            if (lexicalAnalyzer.SymbolType != SymbolType.RParenthesis)
                ErrorMsg("')' is expected.");

            //Get the next symbol from the source file.
            lexicalAnalyzer.GetNextSymbol();
        }

        #endregion

        #region Methods for processing the assembler directives.

        /// <summary>
        /// Processes the DEF assembler directive.
        /// </summary>
        private void ProcessDef()
        {
            string identifier = "";
            long value = 0;

            //Get the identifier from the source file.
            lexicalAnalyzer.GetNextSymbol();
            if (lexicalAnalyzer.SymbolType == SymbolType.Identifier)
            {
                identifier = lexicalAnalyzer.SymbolString;
            }
            else
            {
                if (lexicalAnalyzer.SymbolString.Length == 0)
                    ErrorMsg("identifier is expected.");
                else
                    ErrorMsg("identifier is expected but found '" + lexicalAnalyzer.SymbolString + "'.");
            }

            //Get the numeric constant from the source file.
            lexicalAnalyzer.GetNextSymbol();
            if (lexicalAnalyzer.SymbolType == SymbolType.Number)
            {
                value = lexicalAnalyzer.SymbolValue;

                //Check the value of the constant.
                if ((value < 0) || (value > 255))
                    ErrorMsg("numeric constant is out of range.");
            }
            else
            {
                if (lexicalAnalyzer.SymbolString.Length == 0)
                    ErrorMsg("numeric constant is expected.");
                else
                    ErrorMsg("numeric constant is expected but found '" + lexicalAnalyzer.SymbolString + "'.");
            }

            //Add the identifier to the symbol table.
            if (symbolTable.ContainsKey(identifier))
                ErrorMsg("label or identifier '" + identifier + "' is already defined.");
            symbolTable.Add(identifier, new SymbolTableItem(SymbolType.Identifier, value));

            //Get the next symbol from the source file.
            lexicalAnalyzer.GetNextSymbol();
        }

        #endregion

        #region Methods for processing the assembly instructions.

        /// <summary>
        /// Processes the MOV instructions.
        /// </summary>
        private void ProcessMov()
        {
            int reg1, reg2;
            long constant;

            //Get the next symbol from the source file.
            lexicalAnalyzer.GetNextSymbol();

            //Check whether the first operand is a number or an identifier.
            if ((lexicalAnalyzer.SymbolType == SymbolType.Number) ||
                (lexicalAnalyzer.SymbolType == SymbolType.Identifier))
            {
                //Get the destination data memory address.
                if (lexicalAnalyzer.SymbolType == SymbolType.Number)
                    constant = lexicalAnalyzer.SymbolValue;
                else
                    constant = GetIdentifierValue(lexicalAnalyzer.SymbolString);

                //Get the source register.
                lexicalAnalyzer.GetNextSymbol();
                CheckComma();
                reg1 = GetRegister();

                //Check the data memory address.
                if ((constant < 0) || (constant > 255))
                    ErrorMsg("data memory address is out of range.");

                //MOV d, Ra instruction is written to the program memory.
                WriteProgramMemory(Opcode.MOV_D_R, reg1, (int)constant);
                return;
            }

            //Check whether the current symbol is a '('.
            if (lexicalAnalyzer.SymbolType == SymbolType.LParenthesis)
            {
                //Get the register which stores the indirect address.
                lexicalAnalyzer.GetNextSymbol();
                reg2 = GetRegister();

                //Check the closing parenthesis and ','.
                CheckRParenthesis();
                CheckComma();

                //Get the source data register.
                reg1 = GetRegister();

                //MOV (Rb), Ra instruction is written to the program memory.
                WriteProgramMemory(Opcode.MOV_I_R, 0, reg1, reg2);
                return;
            }

            //The first operand of the other MOV instructions is a register.
            reg1 = GetRegister();
            CheckComma();

            //Check wether the second operand is a number or an identifier.
            if ((lexicalAnalyzer.SymbolType == SymbolType.Number) ||
                (lexicalAnalyzer.SymbolType == SymbolType.Identifier))
            {
                //Get the source data memory address.
                if (lexicalAnalyzer.SymbolType == SymbolType.Number)
                    constant = lexicalAnalyzer.SymbolValue;
                else
                    constant = GetIdentifierValue(lexicalAnalyzer.SymbolString);
                lexicalAnalyzer.GetNextSymbol();

                //Check the data memory address.
                if ((constant < 0) || (constant > 255))
                    ErrorMsg("data memory address is out of range.");

                //MOV Rd, d instruction is written to the program memory.
                WriteProgramMemory(Opcode.MOV_R_D, reg1, (int)constant);
                return;
            }

            //Check whether the current symbol is a '('.
            if (lexicalAnalyzer.SymbolType == SymbolType.LParenthesis)
            {
                //Get the register which stores the indirect address.
                lexicalAnalyzer.GetNextSymbol();
                reg2 = GetRegister();

                //Check the closing parenthesis.
                CheckRParenthesis();

                //MOV Ra, (Rb) instruction is written to the program memory.
                WriteProgramMemory(Opcode.MOV_R_I, 0, reg1, reg2);
                return;
            }

            //Check whether the current symbol is a constant or a '#'.
            if (TryGetConstant(out constant))
            {
                //MOV Rd, #c instruction is written to the program memory.
                WriteProgramMemory(Opcode.MOV_R_C, reg1, (int)constant);
                return;
            }

            //MOV Rd, Ra is a pseudo instruction.
            //OR Rd, Ra, R0 instruction is written to the program memory.
            reg2 = GetRegister();
            WriteProgramMemory(Opcode.OR_R_R_R, reg1, reg2, 0);
        }

        /// <summary>
        /// Processes the ADD and the SUB instructions.
        /// </summary>
        private void ProcessAddSub()
        {
            int rd, ra, rb;
            long constant = 0;
            Keyword keyword = lexicalAnalyzer.Keyword;           

            //Get the destination register.
            lexicalAnalyzer.GetNextSymbol();
            rd = GetRegister();
            CheckComma();

            //Check whether the next symbol is a register.
            if (lexicalAnalyzer.SymbolType == SymbolType.Keyword)
            {
                //Get the source registers.
                ra = GetRegister();
                CheckComma();
                rb = GetRegister();

                if (keyword == Keyword.ADD)
                {
                    //ADD Rd, Ra, Rb instruction is written to the program memory.
                    WriteProgramMemory(Opcode.ADD_R_R_R, rd, ra, rb);
                }
                else
                {
                    //SUB Rd, Ra, Rb instruction is written to the program memory.
                    WriteProgramMemory(Opcode.SUB_R_R_R, rd, ra, rb);
                }
            }
            else
            {
                //Get the constant value.
                constant = GetConstant();

                if (keyword == Keyword.SUB)
                {
                    //SUB Rd, #c is a pseudo instruction.
                    //ADD Rd, #(~c + 1) instruction is written to the program memory.
                    constant = ~constant + 1;
                }
                
                //ADD Rd, #c instruction is written to the program memory.
                WriteProgramMemory(Opcode.ADD_R_C, rd, (int)constant);
            }
        }

        /// <summary>
        /// Processes the AND, OR and XOR instructions.
        /// </summary>
        private void ProcessAndOrXor()
        {
            int rd, ra, rb;
            long constant = 0;
            Keyword keyword = lexicalAnalyzer.Keyword;

            //Get the destination register.
            lexicalAnalyzer.GetNextSymbol();
            rd = GetRegister();
            CheckComma();

            //Check whether the next symbol is a register.
            if (lexicalAnalyzer.SymbolType == SymbolType.Keyword)
            {
                //Get the source registers.
                ra = GetRegister();
                CheckComma();
                rb = GetRegister();

                switch (keyword)
                {
                    case Keyword.OR:
                        //OR Rd, Ra, Rb instruction is written to the program memory.
                        WriteProgramMemory(Opcode.OR_R_R_R, rd, ra, rb);
                        break;
                    case Keyword.XOR:
                        //XOR Rd, Ra, Rb instruction is written to the program memory.
                        WriteProgramMemory(Opcode.XOR_R_R_R, rd, ra, rb);
                        break;
                    default:
                        //AND Rd, Ra, Rb instruction is written to the program memory.
                        WriteProgramMemory(Opcode.AND_R_R_R, rd, ra, rb);
                        break;
                }
            }
            else
            {
                //Get the constant value.
                constant = GetConstant();

                switch (keyword)
                {
                    case Keyword.OR:
                        //OR Rd, #c instruction is written to the program memory.
                        WriteProgramMemory(Opcode.OR_R_C, rd, (int)constant);
                        break;
                    case Keyword.XOR:
                        //XOR Rd, #c instruction is written to the program memory.
                        WriteProgramMemory(Opcode.XOR_R_C, rd, (int)constant);
                        break;
                    default:
                        //AND Rd, #c instruction is written to the program memory.
                        WriteProgramMemory(Opcode.AND_R_C, rd, (int)constant);
                        break;
                }
            }
        }

        /// <summary>
        /// Processes the SL0, SL1, SR0 and SR1 instructions.
        /// </summary>
        private void ProcessShift()
        {
            int rd, ra;
            Keyword keyword = lexicalAnalyzer.Keyword;

            //Get the destination and source registers.
            lexicalAnalyzer.GetNextSymbol();
            rd = GetRegister();
            CheckComma();
            ra = GetRegister();

            switch (keyword)
            {
                case Keyword.SL1:
                    //SL1 Rd, Ra instruction is written to the program memory.
                    WriteProgramMemory(Opcode.SHIFT, rd, ra, (int)ShiftType.SL1);
                    break;
                case Keyword.SR0:
                    //SR0 Rd, Ra instruction is written to the program memory.
                    WriteProgramMemory(Opcode.SHIFT, rd, ra, (int)ShiftType.SR0);
                    break;
                case Keyword.SR1:
                    //SR1 Rd, Ra instruction is written to the program memory.
                    WriteProgramMemory(Opcode.SHIFT, rd, ra, (int)ShiftType.SR1);
                    break;
                case Keyword.ROL:
                    //SR0 Rd, Ra instruction is written to the program memory.
                    WriteProgramMemory(Opcode.SHIFT, rd, ra, (int)ShiftType.ROL);
                    break;
                case Keyword.ROR:
                    //SR1 Rd, Ra instruction is written to the program memory.
                    WriteProgramMemory(Opcode.SHIFT, rd, ra, (int)ShiftType.ROR);
                    break;
                default:
                    //SL0 Rd, Ra instruction is written to the program memory.
                    WriteProgramMemory(Opcode.SHIFT, rd, ra, (int)ShiftType.SL0);
                    break;
            }
        }

        /// <summary>
        /// Processes the program control instructions.
        /// </summary>
        private void ProcessPrgCtrl()
        {
            Keyword keyword = lexicalAnalyzer.Keyword;
            long address = 0;
            int indirectMask = 0;

            //Get the next symbol from the source file.
            lexicalAnalyzer.GetNextSymbol();

            //Get the jump address if the current instruction is not an RTS instruction.
            while (keyword != Keyword.RTS)
            {
                //Use indirect addressing, if the symbol is a '('.
                if (lexicalAnalyzer.SymbolType == SymbolType.LParenthesis)
                {
                    lexicalAnalyzer.GetNextSymbol();
                    address = GetRegister() << 4;
                    CheckRParenthesis();
                    indirectMask = 0x08;
                    break;
                }

                //Use absolute addressing if the simbol is a number.
                if (lexicalAnalyzer.SymbolType == SymbolType.Number)
                {
                    address = lexicalAnalyzer.SymbolValue;
                    lexicalAnalyzer.GetNextSymbol();
                    break;
                }

                //Use absolute addressing if the simbol is an identifier.
                if (lexicalAnalyzer.SymbolType == SymbolType.Identifier)
                {
                    SymbolTableItem item;

                    //If the symbol is a label or an identifier, check the symbol table.
                    if (symbolTable.TryGetValue(lexicalAnalyzer.SymbolString, out item))
                    {
                        address = item.SymbolValue;
                    }
                    else
                    {
                        //Add the label to the unresolved symbols list.
                        unresolvedSymbols.Add(new UnresolvedSymbol(lineNumber, programMemoryAddress, lexicalAnalyzer.SymbolString));
                    }

                    lexicalAnalyzer.GetNextSymbol();
                    break;
                }

                //Invalid label.
                if (lexicalAnalyzer.SymbolString.Length == 0)
                    ErrorMsg("label or identifier is expeced.");
                else
                    ErrorMsg("label or identifier is expected but found '" + lexicalAnalyzer.SymbolString + "'.");

                lexicalAnalyzer.GetNextSymbol();
                break;
            }

            //Check the jump address.
            if ((address < 0) || (address > 255))
                ErrorMsg("jump address is out of range.");

            switch (keyword)
            {
                case Keyword.JZ:
                    //JZ addr/label instruction is written to the program memory.
                    WriteProgramMemory(Opcode.PRG_CTRL, (int)PrgCtrlType.JZ | indirectMask, (int)address);
                    break;
                case Keyword.JNZ:
                    //JNZ addr/label instruction is written to the program memory.
                    WriteProgramMemory(Opcode.PRG_CTRL, (int)PrgCtrlType.JNZ | indirectMask, (int)address);
                    break;
                case Keyword.JC:
                    //JC addr/label instruction is written to the program memory.
                    WriteProgramMemory(Opcode.PRG_CTRL, (int)PrgCtrlType.JC | indirectMask, (int)address);
                    break;
                case Keyword.JNC:
                    //JNC addr/label instruction is written to the program memory.
                    WriteProgramMemory(Opcode.PRG_CTRL, (int)PrgCtrlType.JNC | indirectMask, (int)address);
                    break;
                case Keyword.JSR:
                    //JSR addr/label instruction is written to the program memory.
                    WriteProgramMemory(Opcode.PRG_CTRL, (int)PrgCtrlType.JSR | indirectMask, (int)address);
                    break;
                case Keyword.RTS:
                    //RTS instruction is written to the program memory.
                    WriteProgramMemory(Opcode.PRG_CTRL, (int)PrgCtrlType.RTS, 0);
                    break;
                default:
                    //JMP addr/label instruction is written to the program memory.
                    WriteProgramMemory(Opcode.PRG_CTRL, (int)PrgCtrlType.JMP | indirectMask, (int)address);
                    break;
            }
        }

        /// <summary>
        /// Processes the NOP instruction.
        /// </summary>
        private void ProcessNop()
        {
            //NOP is a pseudo instruction.
            //MOV R0, #0 instruction is written to the program memory.
            WriteProgramMemory(Opcode.MOV_R_C, 0, 0);

            //Get the next symbol from the source file.
            lexicalAnalyzer.GetNextSymbol();
        }

        #endregion

        #region Methods for processing the assembly source file.

        /// <summary>
        /// Processes the assembly source file.
        /// </summary>
        public void Assemble()
        {
            //Check whether the assembly source file is opened.
            if (lexicalAnalyzer.IsOpened == false)
                throw new InvalidOperationException("No assembly source file is opened.");

            //Set the number of errors to zero.
            numOfErrors = 0;

            //First stage: process all of the instructions.
            for (; ; )
            {
                try
                {
                    bool wasLabel = false;

                    //Get the next symbol from the source file.
                    lexicalAnalyzer.GetNextSymbol();

                    //Get the current source code line and the line number.
                    currentLine = lexicalAnalyzer.CurrentLine;
                    lineNumber = lexicalAnalyzer.LineNumber;

                    //Exit from the loop if the end of file has been reached.
                    if (lexicalAnalyzer.SymbolType == SymbolType.EndOfFile)
                        break;

                    //Get the next symbol if the end of line has been reached or the line starts with
                    //a comment. The lexical analyzer reads the next source code line automatically.
                    if ((lexicalAnalyzer.SymbolType == SymbolType.EndOfLine) ||
                        (lexicalAnalyzer.SymbolType == SymbolType.SemiColon))
                        continue;

                    //Check whether the current symbol is a label.
                    if (lexicalAnalyzer.SymbolType == SymbolType.Label)
                    {
                        //A label has been found.
                        wasLabel = true;

                        //Add the label to the symbol table.
                        if (symbolTable.ContainsKey(lexicalAnalyzer.SymbolString))
                            ErrorMsg("label or identifier '" + lexicalAnalyzer.SymbolString + "' is already defined.");
                        symbolTable.Add(lexicalAnalyzer.SymbolString, new SymbolTableItem(SymbolType.Label, programMemoryAddress));

                        //Get the next symbol from the source file.
                        lexicalAnalyzer.GetNextSymbol();
                    }

                    //Get the next symbol if the end of line has been reached or the line starts with
                    //a comment. The lexical analyzer reads the next source code line automatically.
                    if ((lexicalAnalyzer.SymbolType == SymbolType.EndOfLine) ||
                        (lexicalAnalyzer.SymbolType == SymbolType.SemiColon))
                        continue;

                    //The next symbol must be an instruction, which is a keyword.
                    if (lexicalAnalyzer.SymbolType == SymbolType.Keyword)
                    {
                        //Process the current instruction or directive.
                        switch (lexicalAnalyzer.Keyword)
                        {
                            //Process the assembler directives.
                            case Keyword.DEF:
                                if (wasLabel)
                                    ErrorMsg("label is not allowed before directives.");
                                ProcessDef();
                                break;
                            //Process the instructions.
                            case Keyword.MOV:
                                ProcessMov();
                                break;
                            case Keyword.ADD:
                                ProcessAddSub();
                                break;
                            case Keyword.SUB:
                                ProcessAddSub();
                                break;
                            case Keyword.AND:
                                ProcessAndOrXor();
                                break;
                            case Keyword.OR:
                                ProcessAndOrXor();
                                break;
                            case Keyword.XOR:
                                ProcessAndOrXor();
                                break;
                            case Keyword.SL0:
                                ProcessShift();
                                break;
                            case Keyword.SL1:
                                ProcessShift();
                                break;
                            case Keyword.SR0:
                                ProcessShift();
                                break;
                            case Keyword.SR1:
                                ProcessShift();
                                break;
                            case Keyword.ROL:
                                ProcessShift();
                                break;
                            case Keyword.ROR:
                                ProcessShift();
                                break;
                            case Keyword.JMP:
                                ProcessPrgCtrl();
                                break;
                            case Keyword.JZ:
                                ProcessPrgCtrl();
                                break;
                            case Keyword.JNZ:
                                ProcessPrgCtrl();
                                break;
                            case Keyword.JC:
                                ProcessPrgCtrl();
                                break;
                            case Keyword.JNC:
                                ProcessPrgCtrl();
                                break;
                            case Keyword.JSR:
                                ProcessPrgCtrl();
                                break;
                            case Keyword.RTS:
                                ProcessPrgCtrl();
                                break;
                            case Keyword.NOP:
                                ProcessNop();
                                break;
                            default:
                                //Not a valid instruction or directive.
                                ErrorMsg("instruction or directive is expected but found '" + lexicalAnalyzer.SymbolString + "'.");
                                break;
                        }

                        //Check the end of line after the instruction.
                        if ((lexicalAnalyzer.SymbolType == SymbolType.SemiColon) ||
                            (lexicalAnalyzer.SymbolType == SymbolType.EndOfLine))
                            continue;

                        if (lexicalAnalyzer.SymbolString.Length == 0)
                            ErrorMsg("end of line is expected.");
                        else
                            ErrorMsg("end of line is expected but found '" + lexicalAnalyzer.SymbolString + "'.");
                    }
                    else
                    {
                        //Instruction or directive is expected.
                        if (lexicalAnalyzer.SymbolString.Length == 0)
                            ErrorMsg("instruction or directive is expected.");
                        else
                            ErrorMsg("instruction or directive is expected but found '" + lexicalAnalyzer.SymbolString + "'.");
                    }
                }
                catch (LexicalAnalyzerException ex)
                {
                    //Exit from the loop if the number of errors exceeds the maximum value.
                    if (OnError(ex.Message))
                        break;
                    //Read the next line and continue with the processing.
                    lexicalAnalyzer.ReadNextLine();
                }
                catch (AssemblerException ex)
                {
                    //Exit from the loop if the number of errors exceeds the maximum value.
                    if (OnError(ex.Message))
                        break;
                    //Read the next line and continue with the processing.
                    lexicalAnalyzer.ReadNextLine();
                }
            }

            //Stop processing if error ocurred.
            if (numOfErrors != 0)
                return;

            //Second stage: resolve the unresolved symbols and update the program memory.
            foreach (UnresolvedSymbol unresolvedSymbol in unresolvedSymbols)
            {
                try
                {
                    SymbolTableItem item;

                    //Get the constant value from the symbol table.
                    if ((symbolTable.TryGetValue(unresolvedSymbol.Name, out item) == false) ||
                        (item.SymbolType != SymbolType.Label))
                        ErrorMsg(unresolvedSymbol.LineNumber, "unresolved symbol '" + unresolvedSymbol.Name + "'.");

                    //Update the constant value in the program memory.
                    UpdateSymbolValue(unresolvedSymbol, (int)item.SymbolValue);
                }
                catch (AssemblerException ex)
                {
                    //Exit from the loop if the number of errors exceeds the maximum value.
                    if (OnError(ex.Message))
                        break;
                    //Read the next line and continue with the processing.
                    lexicalAnalyzer.ReadNextLine();
                }
            }

            //Stop processing if error ocurred.
            if (numOfErrors != 0)
                return;

            //Third stage: generate the output files.
            GenerateListFile();
            GenerateHexFile();
            GenerateSvfFile();
        }

        #endregion
    }

    /// <summary>
    /// Represents a symbol table item.
    /// </summary>
    internal class SymbolTableItem
    {
        private SymbolType symbolType;
        private long symbolValue;

        /// <summary>
        /// Gets the type of the symbol (identifier or label).
        /// </summary>
        public SymbolType SymbolType { get { return symbolType; } }

        /// <summary>
        /// Gets the value associated to this symbol.
        /// </summary>
        public long SymbolValue { get { return symbolValue; } }

        /// <summary>
        /// Constructor of the SymbolTableItem class.
        /// </summary>
        /// <param name="type">Symbol type.</param>
        /// <param name="value">Value associated to the symbol.</param>
        public SymbolTableItem(SymbolType type, long value)
        {
            symbolType = type;
            symbolValue = value;
        }
    }

    /// <summary>
    /// Contains the necessary data to resolve the unresolved symbol.
    /// </summary>
    internal class UnresolvedSymbol
    {
        private int lineNumber;
        private int programMemoryAddress;
        private string name;

        /// <summary>
        /// Gets the number of the line that contains the unresolved symbol.
        /// </summary>
        public int LineNumber { get { return lineNumber; } }

        /// <summary>
        /// Gets the program memory address of the instruction.
        /// </summary>
        public int ProgramMemoryAddress { get { return programMemoryAddress; } }

        /// <summary>
        /// Gets the name of the symbol.
        /// </summary>
        public string Name { get { return name; } }


        /// <summary>
        /// Constructor of the UnresolvedSymbol class.
        /// </summary>
        /// <param name="lineNumber">Line number.</param>
        /// <param name="programMemoryAddress">Program memory address.</param>
        /// <param name="name">Name of the symbol.</param>
        public UnresolvedSymbol(int lineNumber, int programMemoryAddress, string name)
        {
            this.lineNumber = lineNumber;
            this.programMemoryAddress = programMemoryAddress;
            this.name = name;
        }
    }


    /// <summary>
    /// Exception class used by the assembler.
    /// </summary>
    public class AssemblerException : Exception
    {
        /// <summary>
        /// Constructor of the LexicalAnalyzerException class.
        /// </summary>
        public AssemblerException()
            : base()
        {
        }

        /// <summary>
        /// Constructor of the LexicalAnalyzerException class.
        /// </summary>
        /// <param name="message">Error message.</param>
        public AssemblerException(string message)
            : base(message)
        {
        }

        /// <summary>
        /// Constructor of the LexicalAnalyzerException class.
        /// </summary>
        /// <param name="message">Error message.</param>
        /// <param name="innerException">Inner exception, that is the cause of this exception.</param>
        public AssemblerException(string message, Exception innerException)
            : base(message, innerException)
        {
        }
    }
}
