﻿//******************************************************************************
//* Simple assembler for the MiniRISC CPU.                                     *
//*                                                                            *
//* Written by: Tamas Raikovich                                                *
//* Version   : 2.0                                                            *
//******************************************************************************
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace MiniRISC_as
{
    #region Enumerations.

    /// <summary>
    /// Symbol types.
    /// </summary>
    public enum SymbolType
    {
        //Unknown symbol.
        Unknown = 0,
        //Keyword.
        Keyword,
        //Identifier.
        Identifier,
        //Label (identifier ends with ':').
        Label,
        //Number.
        Number,
        //Constant (number starts with '#').
        Constant,
        //Enf of line indicator.
        EndOfLine,
        //End of file indicator.
        EndOfFile,
        //':' character.
        Colon,
        //';' character.
        SemiColon,
        //',' character.
        Comma,
        //'#' character.
        Sharp,
        //'(' character.
        LParenthesis,
        //')' character.
        RParenthesis
    }

    /// <summary>
    /// Keywords.
    /// </summary>
    public enum Keyword
    {
        //Registers.
        R0 = 0,
        R1,
        R2,
        R3,
        R4,
        R5,
        R6,
        R7,
        R8,
        R9,
        R10,
        R11,
        R12,
        R13,
        R14,
        R15,

        //Instructions.
        MOV,
        ADD,
        SUB,
        AND,
        OR,
        XOR,
        SL0,
        SL1,
        SR0,
        SR1,
        ROL,
        ROR,
        JMP,
        JZ,
        JNZ,
        JC,
        JNC,
        JSR,
        RTS,
        NOP,

        //Directives.
        DEF,

        //Last element of the enumeration (not a valid keyword).
        Last
    }

    #endregion

    /// <summary>
    /// Lexical analyzer.
    /// </summary>
    public class LexicalAnalyzer : IDisposable
    {
        #region Enumerations.

        /// <summary>
        /// Character types.
        /// </summary>
        [Flags]
        private enum CharType : byte
        {
            //Invalid character.
            Invalid = 0x00,
            //Valid character.
            Valid = 0x01,
            //Filler character (space, tab).
            Filler = 0x02,
            //Single character token.
            Token = 0x04,
            //Word terminator character.
            Terminator = 0x08
        }

        #endregion

        //StreamReader object for acessing the assembly source file.
        private StreamReader sourceFile = null;
        
        //Content of the current line.
        private string currentLine = null;
        //Current line number.
        private int lineNumber = 0;
        //Position in the current line.
        private int currentPos = 0;

        //Array that contains the type of each ASCII character.
        private CharType[] charTypeArray = new CharType[256];

        //Dictionary object that contains the keywords.
        private Dictionary<string, Keyword> keywords = new Dictionary<string, Keyword>();

        //Array for storing the identifiers.
        private char[] charArray = new char[64];

        //Type of the current symbol.
        private SymbolType symbolType;
        //Value of the symbol if it is a number.
        private long symbolValue;
        //Value of the symbol if it is a string (identifier, keyword).
        private string symbolString;
        //Current keyword symbol.
        private Keyword keyword;


        #region Properties.

        /// <summary>
        /// Gets a boolean value indicating whether the source file is opened.
        /// </summary>
        public bool IsOpened { get { return (sourceFile != null); } }

        /// <summary>
        /// Gets the current line from the source file.
        /// </summary>
        public string CurrentLine { get { return currentLine; } }

        /// <summary>
        /// Gets the current line number in the source file.
        /// </summary>
        public int LineNumber { get { return lineNumber; } }

        /// <summary>
        /// Gets the character position within the current line.
        /// </summary>
        public int CurrentPos { get { return currentPos; } }

        /// <summary>
        /// Gets the type of the current symbol.
        /// </summary>
        public SymbolType SymbolType { get { return symbolType; } }

        /// <summary>
        /// Gets the value of the current symbol. Valid only if the symbol is a number.
        /// </summary>
        public long SymbolValue { get { return symbolValue; } }

        /// <summary>
        /// Gets the current symbol string. Valid only if the symbol is an identifier.
        /// </summary>
        public string SymbolString { get { return symbolString; } }

        /// <summary>
        /// Gets the current keyword. Valid only if the symbol is a keyword.
        /// </summary>
        public Keyword Keyword { get { return keyword; } }

        /// <summary>
        /// Gets a boolean value indicating whether the current symbol is an instruction.
        /// </summary>
        public bool IsInstruction
        {
            get { return ((symbolType == SymbolType.Keyword) && (keyword >= Keyword.MOV) && (keyword <= Keyword.NOP)); }
        }

        /// <summary>
        /// Gets a boolean value indicating whether the current symbol is a directive.
        /// </summary>
        public bool IsDirective
        {
            get { return ((symbolType == SymbolType.Keyword) && (keyword == Keyword.DEF)); }
        }

        #endregion


        #region Constructor.

        /// <summary>
        /// Constructor of the LexicalAnalyzer class.
        /// </summary>
        public LexicalAnalyzer()
        {
            //Initialize the character type array.
            for (int i = 0; i < charTypeArray.Length; i++)
            {
                //Uppercase and lowercase letters are valid characters.
                if (((i >= (int)'a') && (i <= (int)'z')) || ((i >= (int)'A') && (i <= (int)'Z')))
                {
                    charTypeArray[i] = CharType.Valid;
                    continue;
                }

                //Numbers are valid characters.
                if ((i >= (int)'0') && (i <= (int)'9'))
                {
                    charTypeArray[i] = CharType.Valid;
                    continue;
                }

                //The '_' character is a valid character.
                if (i == (int)'_')
                {
                    charTypeArray[i] = CharType.Valid;
                    continue;
                }

                //Space and tab are filler and terminator characters.
                if ((i == (int)' ') || (i == (int)'\t'))
                {
                    charTypeArray[i] = CharType.Valid | CharType.Filler | CharType.Terminator;
                    continue;
                }

                //The ':', ';', ',', '#', '(' and ')' characters are token and terminator characters.
                if ((i == (int)':') || (i == (int)';') || (i == (int)',') ||
                    (i == (int)'#') || (i == (int)'(') || (i == (int)')'))
                {
                    charTypeArray[i] = CharType.Valid | CharType.Token | CharType.Terminator;
                    continue;
                }

                //Invalid character.
                charTypeArray[i] = CharType.Invalid;
            }

            //Initialize the keyword dictionary.
            for (int i = 0; i < (int)Keyword.Last; i++)
            {
                Keyword keyword = (Keyword)i;
                keywords.Add(keyword.ToString(), keyword);
            }
        }

        #endregion

        #region Methods for opening and closing the assembly source files.

        /// <summary>
        /// Opens the assembly source file.
        /// </summary>
        /// <param name="filename">Name of the file to open.</param>
        public void Open(string filename)
        {
            //Close the previous file if necessary.
            if (sourceFile != null)
                sourceFile.Close();

            //Open the assembly source file.
            try
            {
                sourceFile = new StreamReader(filename, Encoding.ASCII);
            }
            catch
            {
                throw new Exception("Unable to open the assembly source file '" + filename + "'.");
            }

            //Read the first line of the source file.
            lineNumber = 0;
            symbolType = SymbolType.Unknown;
            ReadNextLine();
        }

        /// <summary>
        /// Closes the assembly source file.
        /// </summary>
        public void Close()
        {
            if (sourceFile != null)
            {
                sourceFile.Close();
                sourceFile = null;
            }
        }

        #endregion

        #region IDisposable interface members.

        /// <summary>
        /// Releases the system resources associated with the LexicalAnalyzer object.
        /// </summary>
        void IDisposable.Dispose()
        {
            Close();
        }

        #endregion

        #region Methods for error handling.

        /// <summary>
        /// Raises a LexicalAnalyzerException exception.
        /// </summary>
        /// <param name="msg">Error message.</param>
        private void ErrorMsg(string msg)
        {
            throw new LexicalAnalyzerException("Error at line " + lineNumber.ToString() + ": " + msg);
        }

        #endregion

        #region Methods For checking the character types.

        /// <summary>
        /// Checks whether the character is a valid character.
        /// </summary>
        /// <param name="ch">The character to check.</param>
        /// <returns>True if the character is valid, false otherwise.</returns>
        private bool IsValidCharacter(int ch)
        {
            return ((charTypeArray[ch & 0xff] & CharType.Valid) == CharType.Valid);
        }

        /// <summary>
        /// Checks whether the character is a filler character.
        /// </summary>
        /// <param name="ch">The character to check.</param>
        /// <returns>True if the character is a filler, false otherwise.</returns>
        private bool IsFillerCharacter(int ch)
        {
            return ((charTypeArray[ch & 0xff] & CharType.Filler) == CharType.Filler);
        }

        /// <summary>
        /// Checks whether the character is a token character.
        /// </summary>
        /// <param name="ch">The character to check.</param>
        /// <returns>True if the character is a token, false otherwise.</returns>
        private bool IsTokenCharacter(int ch)
        {
            return ((charTypeArray[ch & 0xff] & CharType.Token) == CharType.Token);
        }

        /// <summary>
        /// Checks whether the character is a terminator character.
        /// </summary>
        /// <param name="ch">The character to check.</param>
        /// <returns>True if the character is a terminator, false otherwise.</returns>
        private bool IsTerminatorCharacter(int ch)
        {
            return ((charTypeArray[ch & 0xff] & CharType.Terminator) == CharType.Terminator);
        }

        #endregion

        #region Methods for parsing numbers (binary, decimal, hexadecimal).

        /// <summary>
        /// Parses a binary number string.
        /// </summary>
        /// <param name="bin">String containing the binary number.</param>
        /// <param name="offset">Starting position in the string.</param>
        /// <param name="length">Length of the binary number in the string.</param>
        /// <returns>Number.</returns>
        private long ParseBinaryNumber(string bin, int offset, int length)
        {
            if (bin == null)
                throw new ArgumentNullException("bin");
            if ((offset < 0) || (offset >= bin.Length))
                throw new ArgumentException("Invalid offset value.", "offset");
            if ((length < 0) || ((offset + length) > bin.Length))
                throw new ArgumentException("Invalid length value.", "length");

            long number = 0;
            long multiplier = 1;

            for (int i = offset + length - 1; i >= offset; i--)
            {
                if ((bin[i] != '0') && (bin[i] != '1'))
                    ErrorMsg("invalid binary number.");

                number += (long)(bin[i] - '0') * multiplier;
                multiplier <<= 1;
            }

            return number;
        }

        /// <summary>
        /// Parses a decimal number string.
        /// </summary>
        /// <param name="dec">String containing the decimal number.</param>
        /// <param name="offset">Starting position in the string.</param>
        /// <param name="length">Length of the decimal number in the string.</param>
        /// <returns>Number.</returns>
        private long ParseDecimalNumber(string dec, int offset, int length)
        {
            if (dec == null)
                throw new ArgumentNullException("dec");
            if ((offset < 0) || (offset >= dec.Length))
                throw new ArgumentException("Invalid offset value.", "offset");
            if ((length < 0) || ((offset + length) > dec.Length))
                throw new ArgumentException("Invalid length value.", "length");

            long number = 0;
            long multiplier = 1;

            for (int i = offset + length - 1; i >= offset; i--)
            {
                if ((dec[i] < '0') || (dec[i] > '9'))
                    ErrorMsg("invalid decimal number.");

                number += (long)(dec[i] - '0') * multiplier;
                multiplier *= 10;
            }

            return number;
        }

        /// <summary>
        /// Parses a hexadecimal number string.
        /// </summary>
        /// <param name="hex">String containing the hexadecimal number.</param>
        /// <param name="offset">Starting position in the string.</param>
        /// <param name="length">Length of the hexadecimal number in the string.</param>
        /// <returns>Number.</returns>
        private long ParseHexadecimalNumber(string hex, int offset, int length)
        {
            if (hex == null)
                throw new ArgumentNullException("dec");
            if ((offset < 0) || (offset >= hex.Length))
                throw new ArgumentException("Invalid offset value.", "offset");
            if ((length < 0) || ((offset + length) > hex.Length))
                throw new ArgumentException("Invalid length value.", "length");

            long number = 0;
            long multiplier = 1;

            for (int i = offset + length - 1; i >= offset; i--)
            {
                if ((hex[i] >= '0') && (hex[i] <= '9'))
                {
                    number += (long)(hex[i] - '0') * multiplier;
                }
                else
                {
                    if ((hex[i] >= 'a') && (hex[i] <= 'f'))
                    {
                        number += (long)(hex[i] - 'a' + 10) * multiplier;
                    }
                    else
                    {
                        if ((hex[i] >= 'A') && (hex[i] <= 'F'))
                        {
                            number += (long)(hex[i] - 'A' + 10) * multiplier;
                        }
                        else
                        {
                            ErrorMsg("invalid hexadecimal number.");
                        }
                    }
                }

                multiplier <<= 4;
            }

            return number;
        }

        /// <summary>
        /// Parses a binary (0b prefix), decimal (no prefix) or hexadecimal (0x prefix) number string.
        /// </summary>
        /// <param name="number">String containing the number.</param>
        /// <param name="offset">Starting position in the string.</param>
        /// <param name="length">Length of the number in the string.</param>
        /// <returns>Number.</returns>
        private long ParseNumber(string number, int offset, int length)
        {
            //Check the binary number prefix (0b).
            if ((number.Length > (offset + 2)) && (number[offset] == '0') && (number[offset + 1] == 'b'))
                return ParseBinaryNumber(number, offset + 2, length - 2);
            //Check the hexadecimal number prefix (0x).
            if ((number.Length > (offset + 2)) && (number[offset] == '0') && (number[offset + 1] == 'x'))
                return ParseHexadecimalNumber(number, offset + 2, length - 2);
            //No prefix: decimal number.
            return ParseDecimalNumber(number, offset, length);
        }

        #endregion

        #region Methods for reading from the source file.

        /// <summary>
        /// Resets the file pointer to the beginning of the source file.
        /// </summary>
        public void ResetFilePointer()
        {
            //Check whether the source file is opened.
            if (sourceFile == null)
                throw new InvalidOperationException("No assembly source file is opened.");

            //Set the file pointer to the beginning of the file and read the first line.
            sourceFile.BaseStream.Seek(0, SeekOrigin.Begin);
            lineNumber = 0;
            symbolType = SymbolType.Unknown;
            ReadNextLine();
        }

        /// <summary>
        /// Reads the next line from the assembly source file.
        /// </summary>
        public void ReadNextLine()
        {
            //Check whether the source file is opened.
            if (sourceFile == null)
                throw new InvalidOperationException("No assembly source file is opened.");

            //Read the next line.
            currentLine = sourceFile.ReadLine();
            symbolType = SymbolType.Unknown;

            //If the end of file has not been reached, increment the line number and reset the position.
            if (currentLine != null)
            {
                lineNumber++;
                currentPos = 0;
            }
        }

        /// <summary>
        /// Calculates the length of the next word in the current line.
        /// </summary>
        /// <returns>Number of characters in the next word.</returns>
        private int CalculateNextWordLength()
        {
            int length = 0;
            int pos = currentPos;

            while (pos < currentLine.Length)
            {
                if (IsTerminatorCharacter((int)currentLine[pos]))
                    break;
                pos++;
                length++;
            }

            return length;
        }

        /// <summary>
        /// Reads the next word from the assembly source file.
        /// </summary>
        /// <returns>String that contains the next word.</returns>
        private string ReadNextWord()
        {
            int length = 0;

            while (currentPos < currentLine.Length)
            {
                if (IsTerminatorCharacter((int)currentLine[currentPos]))
                {
                    break;
                }
                if (IsValidCharacter((int)currentLine[currentPos]) == false)
                {
                    ErrorMsg("invalid character.");
                    break;
                }
                if (length >= charArray.Length)
                {
                    ErrorMsg("identifier is too long.");
                    break;
                }

                charArray[length] = currentLine[currentPos];
                currentPos++;
                length++;
            }

            return new String(charArray, 0, length);
        }

        #endregion

        #region Methods for getting the symbols from the source file.

        /// <summary>
        /// Reads the next symbol from the source file.
        /// </summary>
        /// <param name="skipFillerChars">True, if the filler characters should be skipped before the symbol.</param>
        public void GetNextSymbol(bool skipFillerChars)
        {
            //Check whether the source file is opened.
            if (sourceFile == null)
                throw new InvalidOperationException("No assembly source file is opened.");

            //Read the next line if the previous symbol was a ';' or an end of line.
            if ((symbolType == SymbolType.SemiColon) || (symbolType == SymbolType.EndOfLine))
                ReadNextLine();

            //Set the symbol type to unknown.
            symbolType = SymbolType.Unknown;
            //Set the symbol string to an empty string.
            symbolString = "";
            //Set the symbol value to 0.
            symbolValue = 0;

            //Check whethet the end of file has been reached.
            if (currentLine == null)
            {
                symbolType = SymbolType.EndOfFile;
                return;
            }

            //Skip the filler characters if necessary.
            if (skipFillerChars)
            {
                while (currentPos < currentLine.Length)
                {
                    if (IsFillerCharacter((int)currentLine[currentPos]) == false)
                        break;
                    currentPos++;
                }
            }
            else
            {
                if (IsFillerCharacter((int)currentLine[currentPos]))
                {
                    currentPos++;
                    return;
                }
            }

            //Check whethet the end of line has been reached.
            if (currentPos >= currentLine.Length)
            {
                symbolType = SymbolType.EndOfLine;
                return;
            }

            //Check whether the next character is a token.
            switch (currentLine[currentPos])
            {
                case ':':
                    symbolType = SymbolType.Colon;
                    symbolString = ":";
                    currentPos++;
                    return;
                case ';':
                    symbolType = SymbolType.SemiColon;
                    symbolString = currentLine.Substring(currentPos);
                    return;
                case ',':
                    symbolType = SymbolType.Comma;
                    symbolString = ",";
                    currentPos++;
                    return;
                case '#':
                    //If a number follows the '#', the symbol is a constant.
                    symbolType = SymbolType.Sharp;
                    symbolString = "#";
                    currentPos++;
                    //Return if the end of line has been reached.
                    if (currentPos >= currentLine.Length)
                        return;
                    break;
                case '(':
                    symbolType = SymbolType.LParenthesis;
                    symbolString = "(";
                    currentPos++;
                    return;
                case ')':
                    symbolType = SymbolType.RParenthesis;
                    symbolString = ")";
                    currentPos++;
                    return;
                default:
                    break;
            }

            //If the next character is a numeric digit, read the number.
            if ((currentLine[currentPos] >= '0') && (currentLine[currentPos] <= '9'))
            {
                int length = CalculateNextWordLength();
                int posModifier = (symbolType == SymbolType.Sharp) ? 1 : 0;

                symbolType = (symbolType == SymbolType.Sharp) ? SymbolType.Constant : SymbolType.Number;
                symbolValue = ParseNumber(currentLine, currentPos, length);
                symbolString = currentLine.Substring(currentPos - posModifier, length);
                currentPos += length;
                return;
            }

            //Return if a '#' symbol has been found.
            if (symbolType == SymbolType.Sharp)
                return;

            //If the next character is not a numeric digit, then the next word is
            //an identifier or a keyword. Read the next word from the source file.
            symbolString = ReadNextWord();

            //Check the length of the string.
            if (symbolString.Length == 0)
            {
                symbolType = SymbolType.Unknown;
                return;
            }

            //Check whether the keyword dictionary contains this word.
            if (keywords.TryGetValue(symbolString.ToUpper(), out keyword))
            {
                //The symbol is a keyword.
                symbolType = SymbolType.Keyword;

                //Keywords are not valid label names.
                if ((currentPos < currentLine.Length) && (currentLine[currentPos] == ':'))
                {
                    currentPos++;
                    ErrorMsg("invalid label name '" + symbolString + "'.");
                }
            }
            else
            {
                //The symbol is an identifier.
                symbolType = SymbolType.Identifier;

                //If the next character is a ':', the symbol is a label.
                if ((currentPos < currentLine.Length) && (currentLine[currentPos] == ':'))
                {
                    currentPos++;
                    symbolType = SymbolType.Label;
                }
            }
        }

        /// <summary>
        /// Reads the next symbol from the source file.
        /// </summary>
        public void GetNextSymbol()
        {
            GetNextSymbol(true);
        }

        #endregion
    }


    /// <summary>
    /// Exception class used by the lexical analyzer.
    /// </summary>
    public class LexicalAnalyzerException : Exception
    {
        /// <summary>
        /// Constructor of the LexicalAnalyzerException class.
        /// </summary>
        public LexicalAnalyzerException()
            : base()
        {
        }

        /// <summary>
        /// Constructor of the LexicalAnalyzerException class.
        /// </summary>
        /// <param name="message">Error message.</param>
        public LexicalAnalyzerException(string message)
            : base(message)
        {
        }

        /// <summary>
        /// Constructor of the LexicalAnalyzerException class.
        /// </summary>
        /// <param name="message">Error message.</param>
        /// <param name="innerException">Inner exception, that is the cause of this exception.</param>
        public LexicalAnalyzerException(string message, Exception innerException)
            : base(message, innerException)
        {
        }
    }
}
