//******************************************************************************
//* The control unit of the MiniRISC CPU.                                      *
//******************************************************************************
module control_unit(
   //Clock and reset.
   input  wire        clk,             //Clock signal
   input  wire        rst,             //Reset signal
   
   //Signals related to the program memory.
   output wire [7:0]  prg_mem_addr,    //Address
   input  wire [15:0] prg_mem_din,     //Input data
   
   //Signals related to the data memory.
   output wire        data_mem_wr,     //Write enable signal
   output wire        data_mem_rd,     //Read enable signal
   
   //Constant data from the instruction.
   output wire [7:0]  const_data,      //8-bit constant data
   output wire        alu_shift_in,    //Bit to be shifted in.
   
   //Multiplexer control signals.
   output wire [1:0]  wr_data_sel,     //Write data select for the reg. file
   output wire        addr_op2_sel,    //ALU op. 2 and data mem. addr. select
   
   //Signals related to the register file.
   output wire [3:0]  reg_wr_addr,     //Write address
   output wire        reg_wr_en,       //Write enable signal
   output wire [3:0]  reg_rd_addr_a,   //Read address (A port)
   output wire [3:0]  reg_rd_addr_b,   //Read address (B port)
   
   //Signals related to the ALU.
   output wire [1:0]  alu_op_type,     //Operation type select signal
   output wire        alu_add_sub,     //Arithmetic operation select signal
   output wire [1:0]  alu_logic_op,    //Logic operation select signal
   output wire [1:0]  alu_shift_op,    //Shift operation select signal
   input  wire        alu_flag_z,      //Zero flag
   input  wire        alu_flag_c,      //Carry flag
   
   //Indirect jump address from the register file.
   input  wire [7:0]  ind_jmp_addr,
   
   //Signals for the display unit.
   output wire [7:0]  next_pc,         //Next value of the program counter
   output wire [3:0]  ctrl_state,      //State of the controller (E, D, F, I)
   output wire [15:0] ir_value,        //Value of the instruction register
   output wire        ex_prg_ctrl      //Program control instruction is executed
);

//******************************************************************************
//* Instantiating the program counter (PC).                                    *
//* The program counter drives the address bus of the program memory.          *
//******************************************************************************
wire [7:0] pc_jmp_addr;    //Jump address
wire       pc_clr;         //PC clear signal
wire       pc_jmp;         //PC load signal (jump, subrutine call/return)
wire       pc_rts;         //Return address select (return from subrutine)
wire       pc_jsr;         //Save the return address (subrutine call)
wire       pc_inc;         //PC enable signal

program_counter pc(
   //Clock and reset.
   .clk(clk),              //Clock signal
   .pc_clr(pc_clr),        //Reset signal
   
   //Control signals.
   .pc_jmp(pc_jmp),        //Load signal (jump, subrutine call/return)
   .pc_rts(pc_rts),        //Return address select (return from subrutine)
   .pc_jsr(pc_jsr),        //Save the return address (subrutine call)
   .pc_inc(pc_inc),        //Enable signal
   
   //Jump address.
   .jmp_addr(pc_jmp_addr),
   
   //Value of the program counter.
   .pc(prg_mem_addr),      //Current value of the program counter
   .next_pc(next_pc)       //Next value of the program counter
);


//******************************************************************************
//* Instruction register (IR).                                                 * 
//* The instruction register stores the instruction read from the prog. mem.   *
//* See the 'opcode_defs.vh' file for the instruction format.                  *
//******************************************************************************
reg  [15:0] ir;            //Instruction register.
wire        ir_ld;         //Instruction register load signal

always @(posedge clk)
begin
   if (ir_ld)
      ir <= prg_mem_din;
end

assign ir_value = ir;

//Register write address.
assign reg_wr_addr   = ir[11:8];
//Register read address B.
assign reg_rd_addr_b = ir[3:0];
//Constant data in the instruction.
assign const_data    = ir[7:0];
//New value of the program counter (jump address).
assign pc_jmp_addr   = (ir[11]) ? ind_jmp_addr : ir[7:0];
//Shift operation select signal.
assign alu_shift_op  = {ir[2], ir[0]};
//Bit to be shifted in.
assign alu_shift_in  = ir[1];


//******************************************************************************
//* Instantiating the controller state machine.                                *
//******************************************************************************
controller_fsm controller(
   //Clock and reset.
   .clk(clk),                       //Clock signal
   .rst(rst),                       //Reset signal
   
   //Signals related to the instruction register.
   .ir(ir),                         //Value of the instruction register
   .ir_ld(ir_ld),                   //Instruction register load signal
   
   //Signals related to the program counter.
   .pc_clr(pc_clr),                 //PC clear signal
   .pc_jmp(pc_jmp),                 //PC load signal (jump, subrutine call/return)
   .pc_rts(pc_rts),                 //Return address select (return from subrutine)
   .pc_jsr(pc_jsr),                 //Save the return address (subrutine call)
   .pc_inc(pc_inc),                 //PC increment signal
   
   //Signals related to the datapath.
   .reg_rd_addr_a(reg_rd_addr_a),   //Register read address A
   .wr_data_sel(wr_data_sel),       //Write data select for the reg. file
   .reg_wr_en(reg_wr_en),           //Register file write enable signal
   .addr_op2_sel(addr_op2_sel),     //ALU op. 2 and data mem. addr. select
   .alu_op_type(alu_op_type),       //Operation type select signal
   .alu_add_sub(alu_add_sub),       //Arithmetic operation select signal
   .alu_logic_op(alu_logic_op),     //Logic operation select signal
   .alu_flag_z(alu_flag_z),         //Zero flag
   .alu_flag_c(alu_flag_c),         //Carry flag
   
   //Signals related to the data memory.
   .data_mem_wr(data_mem_wr),       //Write enable signal
   .data_mem_rd(data_mem_rd),       //Read enable signal
   
   //Signals for the display unit.
   .ctrl_state(ctrl_state),         //State of the controller (E, D, F, I)
   .ex_prg_ctrl(ex_prg_ctrl)        //Program control instruction is executed
);


endmodule
