//******************************************************************************
//* MiniRISC CPU                                                               *
//*                                                                            *
//* The controller state machine of the processor.                             *
//******************************************************************************
module controller_fsm(
   //Clock and reset.
   input  wire        clk,                //Clock signal
   input  wire        rst,                //Reset signal
   
   //Signals related to the instruction register.
   input  wire [15:0] ir,                 //Value of the instruction register
   output wire        ir_ld,              //Instruction register load signal
   
   //Signals related to the program counter.
   output wire        pc_clr,             //PC clear signal
   output reg         pc_jmp,             //PC load signal (jump, subrutine call/return)
   output wire        pc_rts,             //Return address select (return from subrutine)
   output wire        pc_jsr,             //Save the return address (subrutine call)
   output wire        pc_inc,             //PC increment signal
   
   //Signals related to the datapath.
   output reg  [3:0]  reg_rd_addr_a,      //Register read address A
   output reg  [1:0]  wr_data_sel,        //Write data select for the reg. file
   output reg         reg_wr_en,          //Register file write enable signal
   output reg         addr_op2_sel,       //ALU op. 2 and data mem. addr. select
   output reg  [1:0]  alu_op_type,        //Operation type select signal
   output wire        alu_add_sub,        //Arithmetic operation select signal
   output reg  [1:0]  alu_logic_op,       //Logic operation select signal
   input  wire        alu_flag_z,         //Zero flag
   input  wire        alu_flag_c,         //Carry flag
   
   //Signals related to the data memory.
   output wire        data_mem_wr,        //Write enable signal
   output wire        data_mem_rd,        //Read enable signal
   
   //Signals for the display unit.
   output wire [3:0]  ctrl_state,         //State of the controller (E, D, F, I)
   output wire        ex_prg_ctrl         //Program control instruction is executed
);

`include "src\control_defs.vh"
`include "src\opcode_defs.vh"

//******************************************************************************
//* Controller state machine.                                                  *
//******************************************************************************
localparam STATE_INIT     = 5'd0;   //Initialization
localparam STATE_FETCH    = 5'd1;   //Instruction fetch
localparam STATE_DECODE   = 5'd2;   //Instruction decode
localparam STATE_EX_LD_A  = 5'd3;   //Instruction execute: load  (absolute)
localparam STATE_EX_LD_I  = 5'd4;   //Instruction execute: load  (indirect)
localparam STATE_EX_ST_A  = 5'd5;   //Instruction execute: store (absolute)
localparam STATE_EX_ST_I  = 5'd6;   //Instruction execute: store (indirect)
localparam STATE_EX_LD_C  = 5'd7;   //Instruction execute: load constant
localparam STATE_EX_ADD_R = 5'd8;   //Instruction execute: ADD (registers)
localparam STATE_EX_ADD_C = 5'd9;   //Instruction execute: ADD (constant)
localparam STATE_EX_SUB_R = 5'd10;  //Instruction execute: SUB (registers)
localparam STATE_EX_AND_R = 5'd11;  //Instruction execute: AND (registers)
localparam STATE_EX_AND_C = 5'd12;  //Instruction execute: AND (constant)
localparam STATE_EX_OR_R  = 5'd13;  //Instruction execute: OR  (registers)
localparam STATE_EX_OR_C  = 5'd14;  //Instruction execute: OR  (constant)
localparam STATE_EX_XOR_R = 5'd15;  //Instruction execute: XOR (registers)
localparam STATE_EX_XOR_C = 5'd16;  //Instruction execute: XOR (constant)
localparam STATE_EX_SHIFT = 5'd17;  //Instruction execute: shift
localparam STATE_EX_CTRL  = 5'd18;  //Instruction execute: program control

//Register that stores the current state.
reg [4:0] current_state;

always @(posedge clk or posedge rst)
begin
   if (rst)
      current_state <= STATE_INIT;
   else
      case (current_state)
         //Initialization.
         STATE_INIT   : current_state <= STATE_FETCH;
         
         //Instruction fetch.
         STATE_FETCH  : current_state <= STATE_DECODE;
         
         //Instruction decode.
         STATE_DECODE : case (ir[15:12])
                           OPCODE_LD_A : current_state <= STATE_EX_LD_A;   //Load  (absolute)
                           OPCODE_LD_I : current_state <= STATE_EX_LD_I;   //Load  (indirect)
                           OPCODE_ST_A : current_state <= STATE_EX_ST_A;   //Store (absolute)
                           OPCODE_ST_I : current_state <= STATE_EX_ST_I;   //Store (indirect)
                           OPCODE_LD_C : current_state <= STATE_EX_LD_C;   //Load constant
                           OPCODE_ADD_R: current_state <= STATE_EX_ADD_R;  //ADD (registers)
                           OPCODE_ADD_C: current_state <= STATE_EX_ADD_C;  //ADD (constant)
                           OPCODE_SUB_R: current_state <= STATE_EX_SUB_R;  //SUB (registers)
                           OPCODE_AND_R: current_state <= STATE_EX_AND_R;  //AND (registers)
                           OPCODE_AND_C: current_state <= STATE_EX_AND_C;  //AND (constant)
                           OPCODE_OR_R : current_state <= STATE_EX_OR_R;   //OR  (registers)
                           OPCODE_OR_C : current_state <= STATE_EX_OR_C;   //OR  (constant)
                           OPCODE_XOR_R: current_state <= STATE_EX_XOR_R;  //XOR (registers)
                           OPCODE_XOR_C: current_state <= STATE_EX_XOR_C;  //XOR (constant)
                           OPCODE_SHIFT: current_state <= STATE_EX_SHIFT;  //Shift
                           OPCODE_CTRL : current_state <= STATE_EX_CTRL;   //Program control
                        endcase
         
         //Instruction execute.
         STATE_EX_LD_A : current_state <= STATE_FETCH;
         STATE_EX_LD_I : current_state <= STATE_FETCH;
         STATE_EX_ST_A : current_state <= STATE_FETCH;
         STATE_EX_ST_I : current_state <= STATE_FETCH;
         STATE_EX_LD_C : current_state <= STATE_FETCH;
         STATE_EX_ADD_R: current_state <= STATE_FETCH;
         STATE_EX_ADD_C: current_state <= STATE_FETCH;
         STATE_EX_SUB_R: current_state <= STATE_FETCH;
         STATE_EX_AND_R: current_state <= STATE_FETCH;
         STATE_EX_AND_C: current_state <= STATE_FETCH;
         STATE_EX_OR_R : current_state <= STATE_FETCH;
         STATE_EX_OR_C : current_state <= STATE_FETCH;
         STATE_EX_XOR_R: current_state <= STATE_FETCH;
         STATE_EX_XOR_C: current_state <= STATE_FETCH;
         STATE_EX_SHIFT: current_state <= STATE_FETCH;
         STATE_EX_CTRL : current_state <= STATE_FETCH;
         
         //Invalid states.
         default       : current_state <= STATE_INIT;
      endcase
end


//******************************************************************************
//* Driving the instruction register control signal.                           *
//******************************************************************************
//The instruction register is loaded in the instruction fetch state.
assign ir_ld = (current_state == STATE_FETCH);


//******************************************************************************
//* Driving the program counter control signals.                               *
//******************************************************************************
//The program counter is cleared in the initialization state.
assign pc_clr = (current_state == STATE_INIT);

//The program counter is incremented in the instruction fetch state.
assign pc_inc = (current_state == STATE_FETCH);

//The program counter is loaded with the new value when the
//jump instructions are executed.
always @(*)
begin
   if (current_state == STATE_EX_CTRL)
      case (ir[10:8])
         PRG_CTRL_JZ : pc_jmp = alu_flag_z;     //Jump if Z is 1
         PRG_CTRL_JNZ: pc_jmp = ~alu_flag_z;    //Jump if Z is 0
         PRG_CTRL_JC : pc_jmp = alu_flag_c;     //Jump if C is 1
         PRG_CTRL_JNC: pc_jmp = ~alu_flag_c;    //Jump if C is 0
         default     : pc_jmp = 1'b1;           //Unconditional jump
      endcase
   else
      pc_jmp = 1'b0;
end

//Save the current program counter value in case of a subrutine call.
assign pc_jsr = (current_state == STATE_EX_CTRL) & (ir[10:8] == PRG_CTRL_JSR);

//Load the program counter with the return address when an RTS instruction is executed.
assign pc_rts = (current_state == STATE_EX_CTRL) & (ir[10:8] == PRG_CTRL_RTS);


//******************************************************************************
//* Driving the internal datapath control signals.                             *
//******************************************************************************
//Register read address (A port).
always @(*)
begin
   case (current_state)
      STATE_EX_ST_A : reg_rd_addr_a = ir[11:8];    //Store (absolute): IR[11:8]
      STATE_EX_ADD_C: reg_rd_addr_a = ir[11:8];    //ADD   (constant): IR[11:8]
      STATE_EX_AND_C: reg_rd_addr_a = ir[11:8];    //AND   (constant): IR[11:8]
      STATE_EX_OR_C : reg_rd_addr_a = ir[11:8];    //OR    (constant): IR[11:8]
      STATE_EX_XOR_C: reg_rd_addr_a = ir[11:8];    //XOR   (constant): IR[11:8]
      default       : reg_rd_addr_a = ir[7:4];     //Other instr.    : IR[7:4]
   endcase
end

//Write data select for the register file.
always @(*)
begin
   case (current_state)
      STATE_EX_LD_A: wr_data_sel = REG_WR_MEM;     //Load (absolute): data memory
      STATE_EX_LD_I: wr_data_sel = REG_WR_MEM;     //Load (indirect): data memory
      STATE_EX_LD_C: wr_data_sel = REG_WR_CNST;    //Load conatant  : constant
      default      : wr_data_sel = REG_WR_ALU;     //Other instr.   : ALU result
   endcase
end

//Register file write enable signal.
always @(*)
begin
   case (current_state)
      STATE_EX_LD_A : reg_wr_en = 1'b1;
      STATE_EX_LD_I : reg_wr_en = 1'b1;
      STATE_EX_LD_C : reg_wr_en = 1'b1;
      STATE_EX_ADD_R: reg_wr_en = 1'b1;
      STATE_EX_ADD_C: reg_wr_en = 1'b1;
      STATE_EX_SUB_R: reg_wr_en = 1'b1;
      STATE_EX_AND_R: reg_wr_en = 1'b1;
      STATE_EX_AND_C: reg_wr_en = 1'b1;
      STATE_EX_OR_R : reg_wr_en = 1'b1;
      STATE_EX_OR_C : reg_wr_en = 1'b1;
      STATE_EX_XOR_R: reg_wr_en = 1'b1;
      STATE_EX_XOR_C: reg_wr_en = 1'b1;
      STATE_EX_SHIFT: reg_wr_en = 1'b1;
      default       : reg_wr_en = 1'b0;
   endcase
end

//ALU operand 2 and data memmemory address select signal.
always @(*)
begin
   case (current_state)
      STATE_EX_LD_A : addr_op2_sel = ALU_OP2_CNST; //Load  (absolute): absolute addr.
      STATE_EX_ST_A : addr_op2_sel = ALU_OP2_CNST; //Store (absolute): absolute addr.
      STATE_EX_ADD_C: addr_op2_sel = ALU_OP2_CNST; //ADD   (constant): constant
      STATE_EX_AND_C: addr_op2_sel = ALU_OP2_CNST; //AND   (constant): constant
      STATE_EX_OR_C : addr_op2_sel = ALU_OP2_CNST; //OR    (constant): constant
      STATE_EX_XOR_C: addr_op2_sel = ALU_OP2_CNST; //XOR   (constant): constant
      default       : addr_op2_sel = ALU_OP2_REGB; //Other instr.    : reg. B / indirect addr.
   endcase
end
   
//ALU operation type select signal.
always @(*)
begin
   case (current_state)
      STATE_EX_ADD_R: alu_op_type = ALU_OP_ARITH;  //ADD (registers): arithmetic op.
      STATE_EX_ADD_C: alu_op_type = ALU_OP_ARITH;  //ADD (constant) : arithmetic op.
      STATE_EX_SUB_R: alu_op_type = ALU_OP_ARITH;  //SUB (registers): arithmetic op.
      STATE_EX_AND_R: alu_op_type = ALU_OP_LOGIC;  //AND (registers): logic op.
      STATE_EX_AND_C: alu_op_type = ALU_OP_LOGIC;  //AND (constant) : logic op.
      STATE_EX_OR_R : alu_op_type = ALU_OP_LOGIC;  //OR  (registers): logic op.
      STATE_EX_OR_C : alu_op_type = ALU_OP_LOGIC;  //OR  (constant) : logic op.
      STATE_EX_XOR_R: alu_op_type = ALU_OP_LOGIC;  //XOR (registers): logic op.
      STATE_EX_XOR_C: alu_op_type = ALU_OP_LOGIC;  //XOR (constant) : logic op.
      STATE_EX_SHIFT: alu_op_type = ALU_OP_SHIFT;  //Shift instr.   : shift op.
      default       : alu_op_type = ALU_OP_MOVE;   //Other instr.   : reg. move op.
   endcase
end

//Arithmetic operation select signal.
assign alu_add_sub = (current_state == STATE_EX_SUB_R);

//Logic operation select signal.
always @(*)
begin
   case (current_state)
      STATE_EX_OR_R : alu_logic_op = LOGIC_OP_OR;
      STATE_EX_OR_C : alu_logic_op = LOGIC_OP_OR;
      STATE_EX_XOR_R: alu_logic_op = LOGIC_OP_XOR;
      STATE_EX_XOR_C: alu_logic_op = LOGIC_OP_XOR;
      default       : alu_logic_op = LOGIC_OP_AND;
   endcase
end


//******************************************************************************
//* Driving the data memory control signals.                                   *
//******************************************************************************
//Data memory write enable signal.
assign data_mem_wr = (current_state == STATE_EX_ST_A) | (current_state == STATE_EX_ST_I);

//Data memory read enable signal.
assign data_mem_rd = (current_state == STATE_EX_LD_A) | (current_state == STATE_EX_LD_I);


//******************************************************************************
//* Driving the signals for the display unit.                                  *
//******************************************************************************
//Current state of the controller.
assign ctrl_state[0] = (current_state == STATE_INIT);       //Init
assign ctrl_state[1] = (current_state == STATE_FETCH);      //Fetch
assign ctrl_state[2] = (current_state == STATE_DECODE);     //Decode
assign ctrl_state[3] = (ctrl_state[2:0] == 3'b000);         //Execute

//Indicates if a program control instruction is executed.
assign ex_prg_ctrl   = (current_state == STATE_EX_CTRL);       


endmodule
