//******************************************************************************
//* MiniRISC CPU                                                               *
//*                                                                            *
//* The datapath of the processor.                                             *
//******************************************************************************
module datapath(
   //Clock signal.
   input  wire       clk,
   
   //Signals related to the data memory.
   output wire [7:0] data_mem_addr,    //Address output
   input  wire [7:0] data_mem_din,     //Input data
   output wire [7:0] data_mem_dout,    //Output data
   
   //Constant data from the instruction.
   input  wire [7:0] const_data_in,    //8-bit constant data
   input  wire       alu_shift_in,     //Bit to be shifted in.
   
   //Multiplexer control signals.
   input  wire [1:0] wr_data_sel,      //Write data select for the reg. file
   input  wire       addr_op2_sel,     //ALU op. 2 and data mem. addr. select
   
   //Signals related to the register file.
   input  wire [3:0] reg_wr_addr,      //Write address
   input  wire       reg_wr_en,        //Write enable signal
   input  wire [3:0] reg_rd_addr_a,    //Read address (A port)
   input  wire [3:0] reg_rd_addr_b,    //Read address (B port)
   
   //Signals related to the ALU.
   input  wire [1:0] alu_op_type,      //Operation type select signal
   input  wire       alu_add_sub,      //Arithmetic operation select signal
   input  wire [1:0] alu_logic_op,     //Logic operation select signal
   input  wire [1:0] alu_shift_op,     //Shift operation select signal
   output wire       alu_flag_z,       //Zero flag
   output wire       alu_flag_c,       //Carry flag
   
   //Signals for the display unit.
   output reg  [7:0] reg_wr_data       //Data to be written into the reg. file
);

`include "src\control_defs.vh"

//******************************************************************************
//* Input data select multiplexer for the register file.                       *
//******************************************************************************
//Result of the ALU operation.
wire [7:0] alu_result;

always @(*)
begin
   case (wr_data_sel)
      REG_WR_ALU : reg_wr_data <= alu_result;         //ALU
      REG_WR_MEM : reg_wr_data <= data_mem_din;       //Data memory
      REG_WR_CNST: reg_wr_data <= const_data_in;      //Constant
      default    : reg_wr_data <= 8'd0;
   endcase
end


//******************************************************************************
//* Instantiating the register file.                                           *
//******************************************************************************
wire [7:0] reg_rd_data_a;
wire [7:0] reg_rd_data_b;

reg_file_1w2r_lut reg_file(
   //Clock signal.
   .clk(clk),
   
   //Write port.
   .wr_addr(reg_wr_addr),        //Register address
   .wr_en(reg_wr_en),            //Write enable signal
   .wr_data(reg_wr_data),        //Input data
   
   //Read port A.
   .rd_addr_a(reg_rd_addr_a),    //Register address
   .rd_data_a(reg_rd_data_a),    //Register data output
   
   //Read port B.
   .rd_addr_b(reg_rd_addr_b),    //Register address
   .rd_data_b(reg_rd_data_b)     //Register data output
);

//Driving the output data bus.
assign data_mem_dout = reg_rd_data_a;


//******************************************************************************
//* ALU operand 2 and data memory address select MUX.                          *
//******************************************************************************
wire [7:0] alu_operand1 = reg_rd_data_a;
reg  [7:0] alu_operand2;

always @(*)
begin
   case (addr_op2_sel)
      ALU_OP2_REGB: alu_operand2 <= reg_rd_data_b;    //Register / indirect addressing
      ALU_OP2_CNST: alu_operand2 <= const_data_in;    //Constant / absolute addressing
   endcase
end

//Driving the data memory address bus.
assign data_mem_addr = alu_operand2;


//******************************************************************************
//* Instantiating the arithmetic and logic unit (ALU).                         *
//******************************************************************************
alu alu(
   //Clock signal.
   .clk(clk),
   
   //Signals related to the function selection.
   .op_type(alu_op_type),        //Operation type select signal
   .add_sub(alu_add_sub),        //Arithmetic operation select signal
   .logic_op(alu_logic_op),      //Logic operation select signal
   .shift_op(alu_shift_op),      //Shift operation select signal
   
   //Input operands.
   .operand1(alu_operand1),      //ALU operand 1
   .operand2(alu_operand2),      //ALU operand 2
   .shift_in(alu_shift_in),      //Bit to be shifted in.
   
   //ALU result.
   .result(alu_result),
   
   //Flags.
   .flag_z(alu_flag_z),          //Zero flag
   .flag_c(alu_flag_c)           //Carry flag
);


endmodule
