//******************************************************************************
//* MiniRISC CPU                                                               *
//*                                                                            *
//* Display unit, which displays the current state of the processor.           *
//******************************************************************************
module display(
   //Clock and reset.
   input  wire        clk,             //Clock signal
   input  wire        rst,             //Reset signal
   
   //Input signals.
   input  wire [3:0]  ctrl_state,      //State of the controller (E, D, F, I)
   input  wire [7:0]  prg_mem_addr,    //Program memory address
   input  wire [15:0] ir_value,        //Value of the instruction register
   input  wire [7:0]  next_pc,         //Next value of the program counter
   input  wire [7:0]  data_mem_addr,   //Data memory address
   input  wire [7:0]  data_mem_dout,   //Data memory output data
   input  wire [7:0]  reg_wr_data,     //Data to be written into the reg. file
   input  wire        ex_load,         //Load instruction is executed
   input  wire        ex_store,        //Store instruction is executed
   input  wire        ex_prg_ctrl,     //Program control instruction is executed
   
   //Display signals.
   output wire [7:0]  seg,             //Segment data
   output reg  [3:0]  dig,             //Digit select signals
   output reg  [4:0]  col              //Column select signals
);

//******************************************************************************
//* Clock divider for generating the ~1 kHz clock enable signal.               *
//******************************************************************************
reg  [13:0] clk_divider;
wire        clk_divider_tc = &clk_divider;

always @(posedge clk)
begin
   if (rst)
      clk_divider <= 14'd0;
   else
      clk_divider <= clk_divider + 14'd1;
end


//******************************************************************************
//* Counter for selecting the display data and generating the digit/column     *
//* select signals.                                                            *
//******************************************************************************
reg [3:0] cnt;

always @(posedge clk)
begin
   if (rst)
      cnt <= 4'd0;
   else
      if (clk_divider_tc)
         if (cnt == 4'd8)
            cnt <= 4'd0;
         else
            cnt <= cnt + 4'd1;
end


//******************************************************************************
//* Generating the digit/column select signals.                                *
//******************************************************************************
always @(*)
begin
   case (cnt)
      4'd1   : {col, dig} = 9'b00000_0010;
      4'd2   : {col, dig} = 9'b00000_0100;
      4'd3   : {col, dig} = 9'b00000_1000;
      4'd4   : {col, dig} = 9'b00001_0000;
      4'd5   : {col, dig} = 9'b00010_0000;
      4'd6   : {col, dig} = 9'b00100_0000;
      4'd7   : {col, dig} = 9'b01000_0000;
      4'd8   : {col, dig} = 9'b10000_0000;
      default: {col, dig} = 9'b00000_0001;
   endcase
end


//******************************************************************************
//* Selecting the dot-matrix display data according to the current CPU state.  *
//******************************************************************************
reg [34:0] dotmatrix_data;

always @(*)
begin
   case (ctrl_state)
      4'b0001: dotmatrix_data = {7'h00, 7'h41, 7'h7f, 7'h41, 7'h00}; //I: init
      4'b0010: dotmatrix_data = {7'h7f, 7'h09, 7'h09, 7'h09, 7'h01}; //F: fetch
      4'b0100: dotmatrix_data = {7'h41, 7'h7f, 7'h41, 7'h41, 7'h3e}; //D: decode
      4'b1000: dotmatrix_data = {7'h7f, 7'h49, 7'h49, 7'h49, 7'h41}; //E: execute
      default: dotmatrix_data = 35'd0;
   endcase
end


//******************************************************************************
//* Selecting the 7-segment display data.                                      *
//******************************************************************************
reg [4:0] sevenseg_digit;

always @(*)
begin
   case (ctrl_state)
      //Fetch: 
      // -leftmost two digits : program memory address,
      // -rightmost two digits: off.
      4'b0010: case (cnt[1:0])
                  2'b10  : sevenseg_digit = {1'b1, prg_mem_addr[3:0]};
                  2'b11  : sevenseg_digit = {1'b1, prg_mem_addr[7:4]};
                  default: sevenseg_digit = 5'd0;
               endcase
               
      //Decode: display the instruction code.
      4'b0100: case (cnt[1:0])
                  2'b00: sevenseg_digit = {1'b1, ir_value[3:0]};
                  2'b01: sevenseg_digit = {1'b1, ir_value[7:4]};
                  2'b10: sevenseg_digit = {1'b1, ir_value[11:8]};
                  2'b11: sevenseg_digit = {1'b1, ir_value[15:12]};
               endcase
      
      //Execute
      4'b1000: case (cnt[1:0])
                  //Rightmost two digits:
                  // -prg. control: off
                  // -store       : data written to the data memory
                  // -other instr.: data written to the reg. file
                  2'b00: if (ex_prg_ctrl)
                            sevenseg_digit = 5'd0;
                         else
                            if (ex_store)
                               sevenseg_digit = {1'b1, data_mem_dout[3:0]};
                            else
                               sevenseg_digit = {1'b1, reg_wr_data[3:0]};
                  2'b01: if (ex_prg_ctrl)
                            sevenseg_digit = 5'd0;
                         else
                            if (ex_store)
                               sevenseg_digit = {1'b1, data_mem_dout[7:4]};
                            else
                               sevenseg_digit = {1'b1, reg_wr_data[7:4]};
      
                  //Leftmost two digits:
                  // -prg. control: next value of the program counter
                  // -load/store  : data memory address
                  // -other instr.: off
                  2'b10: if (ex_prg_ctrl)
                            sevenseg_digit = {1'b1, next_pc[3:0]};
                         else
                            if (ex_store || ex_load)
                               sevenseg_digit = {1'b1, data_mem_addr[3:0]};
                            else
                               sevenseg_digit = 5'd0;
                  2'b11: if (ex_prg_ctrl)
                            sevenseg_digit = {1'b1, next_pc[7:4]};
                         else
                            if (ex_store || ex_load)
                               sevenseg_digit = {1'b1, data_mem_addr[7:4]};
                            else
                               sevenseg_digit = 5'd0;
               endcase
      
      //Init: the seven-segment display is off.
      default: sevenseg_digit = 5'd0;
   endcase
end


//******************************************************************************
//* Seven-segment decoder.                                                     *
//******************************************************************************
reg [6:0] sevenseg_decoder;

always @(*)
begin
   if (sevenseg_digit[4])
      case (sevenseg_digit[3:0])
         4'b0000 : sevenseg_decoder <= 7'b0111111;   // 0
         4'b0001 : sevenseg_decoder <= 7'b0000110;   // 1
         4'b0010 : sevenseg_decoder <= 7'b1011011;   // 2
         4'b0011 : sevenseg_decoder <= 7'b1001111;   // 3
         4'b0100 : sevenseg_decoder <= 7'b1100110;   // 4
         4'b0101 : sevenseg_decoder <= 7'b1101101;   // 5
         4'b0110 : sevenseg_decoder <= 7'b1111101;   // 6
         4'b0111 : sevenseg_decoder <= 7'b0000111;   // 7
         4'b1000 : sevenseg_decoder <= 7'b1111111;   // 8
         4'b1001 : sevenseg_decoder <= 7'b1101111;   // 9
         4'b1010 : sevenseg_decoder <= 7'b1110111;   // A
         4'b1011 : sevenseg_decoder <= 7'b1111100;   // b
         4'b1100 : sevenseg_decoder <= 7'b0111001;   // C
         4'b1101 : sevenseg_decoder <= 7'b1011110;   // d
         4'b1110 : sevenseg_decoder <= 7'b1111001;   // E
         4'b1111 : sevenseg_decoder <= 7'b1110001;   // F
      endcase
   else
      sevenseg_decoder <= 7'b0000000;                // Off
end 


//******************************************************************************
//* Driving the display data output.                                           *
//******************************************************************************
reg [6:0] disp_seg;

always @(*)
begin
   case (cnt)
      4'd4   : disp_seg = dotmatrix_data[6:0];
      4'd5   : disp_seg = dotmatrix_data[13:7];
      4'd6   : disp_seg = dotmatrix_data[20:14];
      4'd7   : disp_seg = dotmatrix_data[27:21];
      4'd8   : disp_seg = dotmatrix_data[34:28];
      default: disp_seg = sevenseg_decoder;
   endcase
end

assign seg = {1'b0, disp_seg};


endmodule
