//******************************************************************************
//* MiniRISC CPU                                                               *
//*                                                                            *
//* Top level module of the processor.                                         *
//******************************************************************************
module minirisc_cpu(
   //Clock and reset.
   input  wire        cpu_clk,         //Clock signal for the processor
   input  wire        disp_clk,        //Clock signal for the display unit
   input  wire        rst,             //Reset signal
   
   //Signals related to the program memory.
   output wire [7:0]  prg_mem_addr,    //Address
   input  wire [15:0] prg_mem_din,     //Input data
   
   //Signals related to the data memory.
   output wire [7:0]  data_mem_addr,   //Address
   output wire        data_mem_wr,     //Write enable signal
   output wire        data_mem_rd,     //Read enable signal
   input  wire [7:0]  data_mem_din,    //Input data
   output wire [7:0]  data_mem_dout,   //Output data
   
   //Signals related to the display unit.
   output wire [7:0]  seg,             //Segment data
   output wire [3:0]  dig,             //Digit select signals
   output wire [4:0]  col              //Column select signals
);

//******************************************************************************
//* Instantiating the control unit.                                            *
//******************************************************************************
wire [7:0]  const_data;
wire        alu_shift_in;
wire [1:0]  wr_data_sel;
wire        addr_op2_sel;
wire [3:0]  reg_wr_addr;
wire        reg_wr_en;
wire [3:0]  reg_rd_addr_a;
wire [3:0]  reg_rd_addr_b;
wire [1:0]  alu_op_type;
wire        alu_add_sub;
wire [1:0]  alu_logic_op;
wire [1:0]  alu_shift_op;
wire        alu_flag_z;
wire        alu_flag_c;
wire [7:0]  next_pc;
wire [3:0]  ctrl_state;
wire [15:0] ir_value;
wire        ex_prg_ctrl;

control_unit control_unit(
   //Clock and reset.
   .clk(cpu_clk),                      //Clock signal
   .rst(rst),                          //Reset signal
   
   //Signals related to the program memory.
   .prg_mem_addr(prg_mem_addr),        //Address
   .prg_mem_din(prg_mem_din),          //Input data
   
   //Signals related to the data memory.
   .data_mem_wr(data_mem_wr),          //Write enable signal
   .data_mem_rd(data_mem_rd),          //Read enable signal
   
   //Constant data from the instruction.
   .const_data(const_data),            //8-bit constant data
   .alu_shift_in(alu_shift_in),        //Bit to be shifted in.
   
   //Multiplexer control signals.
   .wr_data_sel(wr_data_sel),          //Write data select for the reg. file
   .addr_op2_sel(addr_op2_sel),        //ALU op. 2 and data mem. addr. select
   
   //Signals related to the register file.
   .reg_wr_addr(reg_wr_addr),          //Write address
   .reg_wr_en(reg_wr_en),              //Write enable signal
   .reg_rd_addr_a(reg_rd_addr_a),      //Read address (A port)
   .reg_rd_addr_b(reg_rd_addr_b),      //Read address (B port)
   
   //Signals related to the ALU.
   .alu_op_type(alu_op_type),          //Operation type select signal
   .alu_add_sub(alu_add_sub),          //Arithmetic operation select signal
   .alu_logic_op(alu_logic_op),        //Logic operation select signal
   .alu_shift_op(alu_shift_op),        //Shift operation select signal
   .alu_flag_z(alu_flag_z),            //Zero flag
   .alu_flag_c(alu_flag_c),            //Carry flag
   
   //Indirect jump address from the register file.
   .ind_jmp_addr(data_mem_dout),
   
   //Signals for the display unit.
   .next_pc(next_pc),                  //Next value of the program counter
   .ctrl_state(ctrl_state),            //State of the controller (E, D, F, I)
   .ir_value(ir_value),                //Value of the instruction register
   .ex_prg_ctrl(ex_prg_ctrl)           //Program control instruction is executed
);


//******************************************************************************
//* Instantiating the datapath.                                                *
//******************************************************************************
wire [7:0] reg_wr_data;

datapath datapath(
   //Clock signal.
   .clk(cpu_clk),
   
   //Signals related to the data memory.
   .data_mem_addr(data_mem_addr),      //Address output
   .data_mem_din(data_mem_din),        //Input data
   .data_mem_dout(data_mem_dout),      //Output data
   
   //Constant data from the instruction.
   .const_data_in(const_data),         //8-bit constant data
   .alu_shift_in(alu_shift_in),        //Bit to be shifted in.
   
   //Multiplexer control signals.
   .wr_data_sel(wr_data_sel),          //Write data select for the reg. file
   .addr_op2_sel(addr_op2_sel),        //ALU op. 2 and data mem. addr. select
   
   //Signals related to the register file.
   .reg_wr_addr(reg_wr_addr),          //Write address
   .reg_wr_en(reg_wr_en),              //Write enable signal
   .reg_rd_addr_a(reg_rd_addr_a),      //Read address (A port)
   .reg_rd_addr_b(reg_rd_addr_b),      //Read address (B port)
   
   //Signals related to the ALU.
   .alu_op_type(alu_op_type),          //Operation type select signal
   .alu_add_sub(alu_add_sub),          //Arithmetic operation select signal
   .alu_logic_op(alu_logic_op),        //Logic operation select signal
   .alu_shift_op(alu_shift_op),        //Shift operation select signal
   .alu_flag_z(alu_flag_z),            //Zero flag
   .alu_flag_c(alu_flag_c),            //Carry flag
   
   //Signals for the display unit.
   .reg_wr_data(reg_wr_data)           //Data to be written into the reg. file
);


//******************************************************************************
//* Instantiating the display unit.                                            *
//******************************************************************************
display display(
   //Clock and reset.
   .clk(disp_clk),                     //Clock signal
   .rst(rst),                          //Reset signal
   
   //Input signals.
   .ctrl_state(ctrl_state),            //State of the controller (E, D, F, I)
   .prg_mem_addr(prg_mem_addr),        //Program memory address
   .ir_value(ir_value),                //Value of the instruction register
   .next_pc(next_pc),                  //Next value of the program counter
   .data_mem_addr(data_mem_addr),      //Data memory address
   .data_mem_dout(data_mem_dout),      //Data memory output data
   .reg_wr_data(reg_wr_data),          //Data to be written into the reg. file
   .ex_load(data_mem_rd),              //Load instruction is executed
   .ex_store(data_mem_wr),             //Store instruction is executed
   .ex_prg_ctrl(ex_prg_ctrl),          //Program control instruction is executed
   
   //Display signals.
   .seg(seg),                          //Segment data
   .dig(dig),                          //Digit select signals
   .col(col)                           //Column select signals
);


endmodule
