//******************************************************************************
//* MiniRISC system top-level module.                                          *
//******************************************************************************
module minirisc_system(
   //Clock and reset.
   input  wire       clk16M,     //16 MHz clock signal
   input  wire       dev_clk,    //Clock signal from the dev. port
   input  wire       rstbt,      //Reset signal from the reset button
   input  wire       dev_rst,    //Reset signal from the dev. port
   
   //Peripherals.
   input  wire [7:0] sw,         //DIP switch
   input  wire [3:0] bt,         //Push-buttons
   output wire [7:0] ld,         //LEDs
   
   //Poers related to the displays.
   output wire [7:0] seg_n,      //Segment data  (active low)
   output wire [3:0] dig_n,      //Digit select  (active low)
   output wire [4:0] col_n       //Column select (active low)   
);

//******************************************************************************
//* Clock and reset.                                                           *
//******************************************************************************
wire jtag_rst;

wire clk = dev_clk;
wire rst = dev_rst | rstbt | jtag_rst;


//******************************************************************************
//* Instantiating the MiniRISC CPU.                                            *
//******************************************************************************
wire [7:0]  cpu2prgmem_addr;
wire [15:0] prgmem2cpu_data;

wire [7:0]  cpu2datamem_addr;
wire        cpu2datamem_wr;
wire [7:0]  cpu2datamem_data;
reg  [7:0]  datamem2cpu_data;

wire [7:0]  seg;
wire [3:0]  dig;
wire [4:0]  col;

minirisc_cpu minirisc_cpu(
   //Clock and reset.
   .cpu_clk(clk),                      //Clock signal for the processor
   .disp_clk(clk16M),                  //Clock signal for the display unit
   .rst(rst),                          //Reset signal
   
   //Signals related to the program memory.
   .prg_mem_addr(cpu2prgmem_addr),     //Address
   .prg_mem_din(prgmem2cpu_data),      //Input data
   
   //Signals related to the data memory.
   .data_mem_addr(cpu2datamem_addr),   //Address
   .data_mem_wr(cpu2datamem_wr),       //Write enable signal
   .data_mem_rd(),                     //Read enable signal
   .data_mem_din(datamem2cpu_data),    //Input data
   .data_mem_dout(cpu2datamem_data),   //Output data
   
   //Signals related to the display unit.
   .seg(seg),                          //Segment data
   .dig(dig),                          //Digit select signals
   .col(col)                           //Column select signals
);

//Driving the display data and select signals.
assign seg_n = ~seg;
assign dig_n = ~dig;
assign col_n = ~col;


//******************************************************************************
//* Instantiating the program memory.                                          *
//******************************************************************************
prg_memory prg_memory(
   .prg_mem_addr(cpu2prgmem_addr),     //Address input
   .prg_mem_data(prgmem2cpu_data),     //Data output
   .jtag_rst(jtag_rst)                 //Reset output (reset during JTAG update)
);


//******************************************************************************
//* 128 x 8 bit data memory.                                                   *
//* Address range: 0x00 - 0x7F                                                 *
//******************************************************************************
(* ram_style = "distributed" *)
reg  [7:0] data_mem [127:0];

//Data memory address.
wire [6:0] data_mem_addr = cpu2datamem_addr[6:0];
//Data memory write enable signal.
wire       data_mem_wr   = cpu2datamem_wr & ~cpu2datamem_addr[7];
//Input data to the data memory.
wire [7:0] data_mem_din  = cpu2datamem_data;
//Output data from the data memory.
wire [7:0] data_mem_dout = data_mem[data_mem_addr];

always @(posedge clk)
begin
   if (data_mem_wr)
      data_mem[data_mem_addr] <= data_mem_din;
end


//******************************************************************************
//* LED peripheral.                                                            *
//* Address: 0x80 (read/write)                                                 *
//******************************************************************************
reg  [7:0] led_reg;

//Write enable signal for the LED register.
wire       led_reg_wr  = cpu2datamem_wr & cpu2datamem_addr[7];
//Input data for the LED register.
wire [7:0] led_reg_din = cpu2datamem_data;

always @(posedge clk or posedge rst)
begin
   if (rst)
      led_reg <= 8'd0;
   else
      if (led_reg_wr)
         led_reg <= led_reg_din;
end

//Driving the LEDs.
assign ld = led_reg;


//******************************************************************************
//* DIP switch and buttons peripheral.                                         *
//* Address: 0x81: DIP switch (read-only)                                      *
//*          0x82: buttons    (read-only)                                      *
//******************************************************************************
reg [11:0] sw_bt_reg;

//Output data from the DIP switch.
wire [7:0]  sw_dout = sw_bt_reg[7:0];
//Output data from the buttons.
wire [7:0]  bt_dout = {4'd0, sw_bt_reg[11:8]};

always @(posedge clk or posedge rst)
begin
   if (rst)
      sw_bt_reg <= 12'd0;
   else
      sw_bt_reg <= {bt, sw};
end


//******************************************************************************
//* Driving the CPU input data bus.                                            *
//******************************************************************************
always @(*)
begin
   casex (cpu2datamem_addr)
      8'b1xxx_xx00: datamem2cpu_data = led_reg;
      8'b1xxx_xx01: datamem2cpu_data = sw_dout;
      8'b1xxx_xx10: datamem2cpu_data = bt_dout;
      8'b1xxx_xx11: datamem2cpu_data = 8'd0;
      default     : datamem2cpu_data = data_mem_dout;
   endcase
end


endmodule
