//******************************************************************************
//* MiniRISC CPU                                                               *
//*                                                                            *
//* Instruction definitions:                                                   *
//*           15    12 11        8 7      4 3       0                          *
//* -Type A: | opcode |     Rd    |   Ra   |  Rb/sel |                         *
//* -Type B: | opcode | Rd/Ra/sel | address/constant |                         *
//******************************************************************************


//******************************************************************************
//* MOV Rd, d                                                           TYPE B *
//* Load from data memory (absolute addressing): Rd <- DMEM[d]          - -    *
//*                                                                            *
//*   15  12 11     8 7                0                                       *
//*  | 0000 |   Rd   |  data mem. addr. |                                      *
//******************************************************************************
localparam OPCODE_LD_A = 4'b0000;


//******************************************************************************
//* MOV Rd, (Rb)                                                        TYPE A *
//* Load from data memory (indirect addressing): Rd <- DMEM[Rb]         - -    *
//*                                                                            *
//*   15  12 11   8 7    4 3    0                                              *
//*  | 0001 |  Rd  | 0000 |  Rb  |                                             *
//******************************************************************************
localparam OPCODE_LD_I = 4'b0001;


//******************************************************************************
//* MOV d, Ra                                                           TYPE B *
//* Store to data memory (absolute addressing): DMEM[d] <- Ra           - -    *
//*                                                                            *
//*   15  12 11     8 7                0                                       *
//*  | 0010 |   Ra   |  data mem. addr. |                                      *
//******************************************************************************
localparam OPCODE_ST_A = 4'b0010;


//******************************************************************************
//* MOV (Rb), Ra                                                        TYPE A *
//* Store to data memory (indirect addressing): DMEM[Rb] <- Ra          - -    *
//*                                                                            *
//*   15  12 11   8 7    4 3    0                                              *
//*  | 0011 | 0000 |  Ra  |  Rb  |                                             *
//******************************************************************************
localparam OPCODE_ST_I = 4'b0011;


//******************************************************************************
//* MOV Rd, #c                                                          TYPE B *
//* Load constant: Rd <- c                                              - -    *
//*                                                                            *
//*   15  12 11     8 7                0                                       *
//*  | 0100 |   Rd   |     constant     |                                      *
//******************************************************************************
localparam OPCODE_LD_C = 4'b0100;


//******************************************************************************
//* ADD Rd, Ra, Rb                                                      TYPE A *
//* Addition (registers): Rd <- Ra + Rb                                 C Z    *
//*                                                                            *
//*   15  12 11   8 7    4 3    0                                              *
//*  | 0101 |  Rd  |  Ra  |  Rb  |                                             *
//******************************************************************************
localparam OPCODE_ADD_R = 4'b0101;


//******************************************************************************
//* ADD Rd, #c                                                          TYPE B *
//* Addition (constant): Rd <- Rd + c                                   C Z    *
//*                                                                            *
//*   15  12 11     8 7                0                                       *
//*  | 0110 |   Rd   |     constant     |                                      *
//******************************************************************************
localparam OPCODE_ADD_C = 4'b0110;


//******************************************************************************
//* SUB Rd, Ra, Rb                                                      TYPE A *
//* Subtraction (registers): Rd <- Ra - Rb                              C Z    *
//*                                                                            *
//*   15  12 11   8 7    4 3    0                                              *
//*  | 0111 |  Rd  |  Ra  |  Rb  |                                             *
//******************************************************************************
localparam OPCODE_SUB_R = 4'b0111;


//******************************************************************************
//* AND Rd, Ra, Rb                                                      TYPE A *
//* Bitwise AND (registers): Rd <- Ra & Rb                              - Z    *
//*                                                                            *
//*   15  12 11   8 7    4 3    0                                              *
//*  | 1000 |  Rd  |  Ra  |  Rb  |                                             *
//******************************************************************************
localparam OPCODE_AND_R = 4'b1000;


//******************************************************************************
//* AND Rd, #c                                                          TYPE B *
//* Bitwise AND (constant): Rd <- Rd & c                                - Z    *
//*                                                                            *
//*   15  12 11     8 7                0                                       *
//*  | 1001 |   Rd   |     constant     |                                      *
//******************************************************************************
localparam OPCODE_AND_C = 4'b1001;


//******************************************************************************
//* OR Rd, Ra, Rb                                                       TYPE A *
//* Bitwise OR (registers): Rd <- Ra | Rb                               - Z    *
//*                                                                            *
//*   15  12 11   8 7    4 3    0                                              *
//*  | 1010 |  Rd  |  Ra  |  Rb  |                                             *
//******************************************************************************
localparam OPCODE_OR_R = 4'b1010;


//******************************************************************************
//* OR Rd, #c                                                           TYPE B *
//* Bitwise OR (constant): Rd <- Rd | c                                 - Z    *
//*                                                                            *
//*   15  12 11     8 7                0                                       *
//*  | 1011 |   Rd   |     constant     |                                      *
//******************************************************************************
localparam OPCODE_OR_C = 4'b1011;


//******************************************************************************
//* XOR Rd, Ra, Rb                                                      TYPE A *
//* Bitwise XOR (registers): Rd <- Ra ^ Rb                              - Z    *
//*                                                                            *
//*   15  12 11   8 7    4 3    0                                              *
//*  | 1100 |  Rd  |  Ra  |  Rb  |                                             *
//******************************************************************************
localparam OPCODE_XOR_R = 4'b1100;


//******************************************************************************
//* XOR Rd, #c                                                          TYPE B *
//* Bitwise XOR (constant): Rd <- Rd ^ c                                - Z    *
//*                                                                            *
//*   15  12 11     8 7                0                                       *
//*  | 1101 |   Rd   |     constant     |                                      *
//******************************************************************************
localparam OPCODE_XOR_C = 4'b1101;


//******************************************************************************
//* SL0 Rd, Ra                                                          TYPE A *
//* Shift left  (0 shifted in): Rd <- {Ra[6:0], 0}, C <- Ra[7]          C Z    *
//* SL1 Rd, Ra                                                                 *
//* Shift left  (1 shifted in): Rd <- {Ra[6:0], 1}, C <- Ra[7]                 *
//* SR0 Rd, Ra                                                                 *
//* Shift right (0 shifted in): Rd <- {0, Ra[7:1]}, C <- Ra[0]                 *
//* SR1 Rd, Ra                                                                 *
//* Shift right (1 shifted in): Rd <- {1, Ra[7:1]}, C <- Ra[0]                 *
//* ROL Rd, Ra                                                                 *
//* Rotate register left      : Rd <- {Ra[6:0], Ra[7]}, C <- Ra[7]             *
//* ROR Rd, Ra                                                                 *
//* Rotate register right     : Rd <- {Ra[0], Ra[7:1]}, C <- Ra[0]             *
//*                                                                            *
//*   15  12 11   8 7    4 3    0                                              *
//*  | 1110 |  Rd  |  Ra  | 0RID |                                             *
//*                          |||                                               *
//*                          ||\-----> Direction: left (0) / right (1)         *
//*                          |\------> Bit to be shifted in                    *
//*                          \-------> Operation: rotate (1) / shift (0)       *
//******************************************************************************
localparam OPCODE_SHIFT = 4'b1110;

localparam SHIFT_SHL  = 2'b00;
localparam SHIFT_SHR  = 2'b01;
localparam SHIFT_ROL  = 2'b10;
localparam SHIFT_ROR  = 2'b11;


//******************************************************************************
//* JMP address/label/(Ra) - Unconditional jump                         TYPE B *
//* JZ  address/label/(Ra) - Jump if Z flag is 1                        - -    *
//* JNZ address/label/(Ra) - Jump if Z flag is 0                               *
//* JC  address/label/(Ra) - Jump if C flag is 1                               *
//* JNC address/label/(Ra) - Jump if C flag is 0                               *
//* JSR address/label/(Ra) - Subrutine call                                    *
//* RTS                    - Return from subrutine                             *
//*                                                                            *
//*   15  12 11 10  8 7      4 3       0                                       *
//*  | 1111 | 0| op. | prog. mem. addr. | - PC <- prog. mem. addr.             *
//*  | 1111 | 1| op. |   Ra   |  0000   | - PC <- Ra                           *
//*           |   |                                                            *
//*           |   \------> Operation : see the PRG_CTRL_xxx constants          *
//*           \----------> Addressing: absolute (0) / indirect (1)             *
//******************************************************************************
localparam OPCODE_CTRL  = 4'b1111;

localparam PRG_CTRL_JMP = 3'b000;
localparam PRG_CTRL_JZ  = 3'b001;
localparam PRG_CTRL_JNZ = 3'b010;
localparam PRG_CTRL_JC  = 3'b011;
localparam PRG_CTRL_JNC = 3'b100;
localparam PRG_CTRL_JSR = 3'b101;
localparam PRG_CTRL_RTS = 3'b110;

