//******************************************************************************
//* MiniRISC CPU                                                               *
//*                                                                            *
//* 16 x 8 bit register file with 1 write and 2 read ports. R0 is always 0.    *
//* Normal registers are used in this version.                                 *
//******************************************************************************
module reg_file_1w2r(
   //Clock signal.
   input  wire       clk,
   
   //Write port.
   input  wire [3:0] wr_addr,       //Register address
   input  wire       wr_en,         //Write enable signal
   input  wire [7:0] wr_data,       //Input data
   
   //Read port A.
   input  wire [3:0] rd_addr_a,     //Register address
   output reg  [7:0] rd_data_a,     //Register data output
   
   //Read port B.
   input  wire [3:0] rd_addr_b,     //Register address
   output reg  [7:0] rd_data_b      //Register data output
);

//******************************************************************************
//* Declaring 16 eight-bit registers.                                          *
//******************************************************************************
reg [7:0] r1, r2, r3, r4, r5, r6, r7;
reg [7:0] r8, r9, r10, r11, r12, r13, r14, r15;


//******************************************************************************
//* Implementing the write port. Write operations are synchronous.             *
//* The value of R0 is always 0 so writes to R0 are ignored.                   *
//******************************************************************************
always @(posedge clk)
begin
   if (wr_en)
      case (wr_addr)
         4'd1 : r1  <= wr_data;
         4'd2 : r2  <= wr_data;
         4'd3 : r3  <= wr_data;
         4'd4 : r4  <= wr_data;
         4'd5 : r5  <= wr_data;
         4'd6 : r6  <= wr_data;
         4'd7 : r7  <= wr_data;
         4'd8 : r8  <= wr_data;
         4'd9 : r9  <= wr_data;
         4'd10: r10 <= wr_data;
         4'd11: r11 <= wr_data;
         4'd12: r12 <= wr_data;
         4'd13: r13 <= wr_data;
         4'd14: r14 <= wr_data;
         4'd15: r15 <= wr_data;
      endcase
end


//******************************************************************************
//* Implementing the read port A. Read operation is asynchronous, a 16 to 1    *
//* multiplexer selects between the register outputs. R0 is always 0.          *
//******************************************************************************
always @(*)
begin
   case (rd_addr_a)
      4'd0 : rd_data_a <= 8'd0;
      4'd1 : rd_data_a <= r1;
      4'd2 : rd_data_a <= r2;
      4'd3 : rd_data_a <= r3;
      4'd4 : rd_data_a <= r4;
      4'd5 : rd_data_a <= r5;
      4'd6 : rd_data_a <= r6;
      4'd7 : rd_data_a <= r7;
      4'd8 : rd_data_a <= r8;
      4'd9 : rd_data_a <= r9;
      4'd10: rd_data_a <= r10;
      4'd11: rd_data_a <= r11;
      4'd12: rd_data_a <= r12;
      4'd13: rd_data_a <= r13;
      4'd14: rd_data_a <= r14;
      4'd15: rd_data_a <= r15;
   endcase
end


//******************************************************************************
//* Implementing the read port B. Read operation is asynchronous, a 16 to 1    *
//* multiplexer selects between the register outputs. R0 is always 0.          *
//******************************************************************************
always @(*)
begin
   case (rd_addr_b)
      4'd0 : rd_data_b <= 8'd0;
      4'd1 : rd_data_b <= r1;
      4'd2 : rd_data_b <= r2;
      4'd3 : rd_data_b <= r3;
      4'd4 : rd_data_b <= r4;
      4'd5 : rd_data_b <= r5;
      4'd6 : rd_data_b <= r6;
      4'd7 : rd_data_b <= r7;
      4'd8 : rd_data_b <= r8;
      4'd9 : rd_data_b <= r9;
      4'd10: rd_data_b <= r10;
      4'd11: rd_data_b <= r11;
      4'd12: rd_data_b <= r12;
      4'd13: rd_data_b <= r13;
      4'd14: rd_data_b <= r14;
      4'd15: rd_data_b <= r15;
   endcase
end


endmodule
